@@ -3,7 +3,6 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
-#include <assert.h>
 #include <errno.h>
 
 #if defined(_DEBUG)
@@ -15,7 +14,9 @@
 #define LOG_MODULE "csi"
 #define LOG_ENABLE_DBG 0
 #include "log.h"
+#include "char32.h"
 #include "config.h"
+#include "debug.h"
 #include "grid.h"
 #include "selection.h"
 #include "sixel.h"
@@ -32,8 +33,6 @@ static void
 sgr_reset(struct terminal *term)
 {
     memset(&term->vt.attrs, 0, sizeof(term->vt.attrs));
-    term->vt.attrs.fg = term->colors.fg;
-    term->vt.attrs.bg = term->colors.bg;
 }
 
 static const char *
@@ -115,8 +114,8 @@ csi_sgr(struct terminal *term)
         case 35:
         case 36:
         case 37:
-            term->vt.attrs.have_fg = 1;
-            term->vt.attrs.fg = term->colors.table[param - 30];
+            term->vt.attrs.fg_src = COLOR_BASE16;
+            term->vt.attrs.fg = param - 30;
             break;
 
         case 38: {
@@ -124,9 +123,8 @@ csi_sgr(struct terminal *term)
             if (term->vt.params.idx - i - 1 >= 2 &&
                 term->vt.params.v[i + 1].value == 5)
             {
-                uint8_t idx = term->vt.params.v[i + 2].value;
-                term->vt.attrs.have_fg = 1;
-                term->vt.attrs.fg = term->colors.table[idx];
+                term->vt.attrs.fg_src = COLOR_BASE256;
+                term->vt.attrs.fg = term->vt.params.v[i + 2].value;
                 i += 2;
 
             }
@@ -138,7 +136,7 @@ csi_sgr(struct terminal *term)
                 uint8_t r = term->vt.params.v[i + 2].value;
                 uint8_t g = term->vt.params.v[i + 3].value;
                 uint8_t b = term->vt.params.v[i + 4].value;
-                term->vt.attrs.have_fg = 1;
+                term->vt.attrs.fg_src = COLOR_RGB;
                 term->vt.attrs.fg = r << 16 | g << 8 | b;
                 i += 4;
             }
@@ -149,9 +147,8 @@ csi_sgr(struct terminal *term)
             {
                 const struct vt_param *param = &term->vt.params.v[i];
 
-                uint8_t idx = param->sub.value[1];
-                term->vt.attrs.have_fg = 1;
-                term->vt.attrs.fg = term->colors.table[idx];
+                term->vt.attrs.fg_src = COLOR_BASE256;
+                term->vt.attrs.fg = param->sub.value[1];
             }
 
             /*
@@ -180,7 +177,7 @@ csi_sgr(struct terminal *term)
                 uint8_t g = param->sub.value[g_idx];
                 uint8_t b = param->sub.value[b_idx];
 
-                term->vt.attrs.have_fg = 1;
+                term->vt.attrs.fg_src = COLOR_RGB;
                 term->vt.attrs.fg = r << 16 | g << 8 | b;
             }
 
@@ -196,7 +193,7 @@ csi_sgr(struct terminal *term)
         }
 
         case 39:
-            term->vt.attrs.have_fg = 0;
+            term->vt.attrs.fg_src = COLOR_DEFAULT;
             break;
 
         /* Regular background colors */
@@ -208,8 +205,8 @@ csi_sgr(struct terminal *term)
         case 45:
         case 46:
         case 47:
-            term->vt.attrs.have_bg = 1;
-            term->vt.attrs.bg = term->colors.table[param - 40];
+            term->vt.attrs.bg_src = COLOR_BASE16;
+            term->vt.attrs.bg = param - 40;
             break;
 
         case 48: {
@@ -217,9 +214,8 @@ csi_sgr(struct terminal *term)
             if (term->vt.params.idx - i - 1 >= 2 &&
                 term->vt.params.v[i + 1].value == 5)
             {
-                uint8_t idx = term->vt.params.v[i + 2].value;
-                term->vt.attrs.have_bg = 1;
-                term->vt.attrs.bg = term->colors.table[idx];
+                term->vt.attrs.bg_src = COLOR_BASE256;
+                term->vt.attrs.bg = term->vt.params.v[i + 2].value;
                 i += 2;
 
             }
@@ -231,7 +227,7 @@ csi_sgr(struct terminal *term)
                 uint8_t r = term->vt.params.v[i + 2].value;
                 uint8_t g = term->vt.params.v[i + 3].value;
                 uint8_t b = term->vt.params.v[i + 4].value;
-                term->vt.attrs.have_bg = 1;
+                term->vt.attrs.bg_src = COLOR_RGB;
                 term->vt.attrs.bg = r << 16 | g << 8 | b;
                 i += 4;
             }
@@ -242,9 +238,8 @@ csi_sgr(struct terminal *term)
             {
                 const struct vt_param *param = &term->vt.params.v[i];
 
-                uint8_t idx = param->sub.value[1];
-                term->vt.attrs.have_bg = 1;
-                term->vt.attrs.bg = term->colors.table[idx];
+                term->vt.attrs.bg_src = COLOR_BASE256;
+                term->vt.attrs.bg = param->sub.value[1];
             }
 
             /*
@@ -273,7 +268,7 @@ csi_sgr(struct terminal *term)
                 uint8_t g = param->sub.value[g_idx];
                 uint8_t b = param->sub.value[b_idx];
 
-                term->vt.attrs.have_bg = 1;
+                term->vt.attrs.bg_src = COLOR_RGB;
                 term->vt.attrs.bg = r << 16 | g << 8 | b;
             }
 
@@ -287,7 +282,7 @@ csi_sgr(struct terminal *term)
             break;
         }
         case 49:
-            term->vt.attrs.have_bg = 0;
+            term->vt.attrs.bg_src = COLOR_DEFAULT;
             break;
 
         /* Bright foreground colors */
@@ -299,8 +294,8 @@ csi_sgr(struct terminal *term)
         case 95:
         case 96:
         case 97:
-            term->vt.attrs.have_fg = 1;
-            term->vt.attrs.fg = term->colors.table[param - 90 + 8];
+            term->vt.attrs.fg_src = COLOR_BASE16;
+            term->vt.attrs.fg = param - 90 + 8;
             break;
 
         /* Bright background colors */
@@ -312,8 +307,8 @@ csi_sgr(struct terminal *term)
         case 105:
         case 106:
         case 107:
-            term->vt.attrs.have_bg = 1;
-            term->vt.attrs.bg = term->colors.table[param - 100 + 8];
+            term->vt.attrs.bg_src = COLOR_BASE16;
+            term->vt.attrs.bg = param - 100 + 8;
             break;
 
         default:
@@ -345,10 +340,7 @@ decset_decrst(struct terminal *term, unsigned param, bool enable)
         if (enable)
             LOG_WARN("unimplemented: 132 column mode (DECCOLM)");
 
-        term_erase(
-            term,
-            &(struct coord){0, 0},
-            &(struct coord){term->cols - 1, term->rows - 1});
+        term_erase(term, 0, 0, term->rows - 1, term->cols - 1);
         term_cursor_home(term);
         break;
 
@@ -401,6 +393,10 @@ decset_decrst(struct terminal *term, unsigned param, bool enable)
         term->reverse_wrap = enable;
         break;
 
+    case 80:
+        term->sixel.scrolling = !enable;
+        break;
+
     case 1000:
         if (enable)
             term->mouse_tracking = MOUSE_CLICK;
@@ -461,6 +457,13 @@ decset_decrst(struct terminal *term, unsigned param, bool enable)
             term->mouse_reporting = MOUSE_NORMAL;
         break;
 
+    case 1016:
+        if (enable)
+            term->mouse_reporting = MOUSE_SGR_PIXELS;
+        else if (term->mouse_reporting == MOUSE_SGR_PIXELS)
+            term->mouse_reporting = MOUSE_NORMAL;
+        break;
+
     case 1034:
         /* smm */
         LOG_DBG("%s 8-bit meta mode", enable ? "enabling" : "disabling");
@@ -489,23 +492,32 @@ decset_decrst(struct terminal *term, unsigned param, bool enable)
         break;
 #endif
 
+    case 1048:
+        if (enable)
+            term_save_cursor(term);
+        else
+            term_restore_cursor(term, &term->grid->saved_cursor);
+        break;
+
+    case 47:
+    case 1047:
     case 1049:
         if (enable && term->grid != &term->alt) {
             selection_cancel(term);
 
+            if (param == 1049)
+                term_save_cursor(term);
+
             term->grid = &term->alt;
 
+            /* Cursor retains its position from the normal grid */
             term_cursor_to(
                 term,
-                min(term->grid->cursor.point.row, term->rows - 1),
-                min(term->grid->cursor.point.col, term->cols - 1));
+                min(term->normal.cursor.point.row, term->rows - 1),
+                min(term->normal.cursor.point.col, term->cols - 1));
 
-            tll_free(term->alt.scroll_damage);
-
-            term_erase(
-                term,
-                &(struct coord){0, 0},
-                &(struct coord){term->cols - 1, term->rows - 1});
+            tll_free(term->normal.scroll_damage);
+            term_erase(term, 0, 0, term->rows - 1, term->cols - 1);
         }
 
         else if (!enable && term->grid == &term->alt) {
@@ -513,12 +525,13 @@ decset_decrst(struct terminal *term, unsigned param, bool enable)
 
             term->grid = &term->normal;
 
+            /* Cursor retains its position from the alt grid */
             term_cursor_to(
-                term,
-                min(term->grid->cursor.point.row, term->rows - 1),
-                min(term->grid->cursor.point.col, term->cols - 1));
+                term, min(term->alt.cursor.point.row, term->rows - 1),
+                min(term->alt.cursor.point.col, term->cols - 1));
 
-            tll_free(term->alt.scroll_damage);
+            if (param == 1049)
+                term_restore_cursor(term, &term->grid->saved_cursor);
 
             /* Delete all sixel images on the alt screen */
             tll_foreach(term->alt.sixel_images, it) {
@@ -526,14 +539,31 @@ decset_decrst(struct terminal *term, unsigned param, bool enable)
                 tll_remove(term->alt.sixel_images, it);
             }
 
+            tll_free(term->alt.scroll_damage);
             term_damage_all(term);
         }
+        term_update_ascii_printer(term);
+        break;
+
+    case 1070:
+        term->sixel.use_private_palette = enable;
         break;
 
     case 2004:
         term->bracketed_paste = enable;
         break;
 
+    case 2026:
+        if (enable)
+            term_enable_app_sync_updates(term);
+        else
+            term_disable_app_sync_updates(term);
+        break;
+
+    case 8452:
+        term->sixel.cursor_right_of_graphics = enable;
+        break;
+
     case 27127:
         term->modify_escape_key = enable;
         break;
@@ -577,6 +607,7 @@ decrqm(const struct terminal *term, unsigned param, bool *enabled)
     case 12: *enabled = term->cursor_blink.decset; return true;
     case 25: *enabled = !term->hide_cursor; return true;
     case 45: *enabled = term->reverse_wrap; return true;
+    case 80: *enabled = !term->sixel.scrolling; return true;
     case 1000: *enabled = term->mouse_tracking == MOUSE_CLICK; return true;
     case 1001: *enabled = false; return true;
     case 1002: *enabled = term->mouse_tracking == MOUSE_DRAG; return true;
@@ -586,12 +617,18 @@ decrqm(const struct terminal *term, unsigned param, bool *enabled)
     case 1006: *enabled = term->mouse_reporting == MOUSE_SGR; return true;
     case 1007: *enabled = term->alt_scrolling; return true;
     case 1015: *enabled = term->mouse_reporting == MOUSE_URXVT; return true;
+    case 1016: *enabled = term->mouse_reporting == MOUSE_SGR_PIXELS; return true;
     case 1034: *enabled = term->meta.eight_bit; return true;
     case 1035: *enabled = term->num_lock_modifier; return true;
     case 1036: *enabled = term->meta.esc_prefix; return true;
     case 1042: *enabled = term->bell_action_enabled; return true;
+    case 47:   /* FALLTHROUGH */
+    case 1047: /* FALLTHROUGH */
     case 1049: *enabled = term->grid == &term->alt; return true;
+    case 1079: *enabled = term->sixel.use_private_palette; return true;
     case 2004: *enabled = term->bracketed_paste; return true;
+    case 2026: *enabled = term->render.app_sync_updates.enabled; return true;
+    case 8452: *enabled = term->sixel.cursor_right_of_graphics; return true;
     case 27127: *enabled = term->modify_escape_key; return true;
     case 737769: *enabled = term_ime_is_enabled(term); return true;
     }
@@ -613,6 +650,8 @@ xtsave(struct terminal *term, unsigned param)
     case 12: term->xtsave.cursor_blink = term->cursor_blink.decset; break;
     case 25: term->xtsave.show_cursor = !term->hide_cursor; break;
     case 45: term->xtsave.reverse_wrap = term->reverse_wrap; break;
+    case 47: term->xtsave.alt_screen = term->grid == &term->alt; break;
+    case 80: term->xtsave.sixel_display_mode = !term->sixel.scrolling; break;
     case 1000: term->xtsave.mouse_click = term->mouse_tracking == MOUSE_CLICK; break;
     case 1001: break;
     case 1002: term->xtsave.mouse_drag = term->mouse_tracking == MOUSE_DRAG; break;
@@ -622,12 +661,18 @@ xtsave(struct terminal *term, unsigned param)
     case 1006: term->xtsave.mouse_sgr = term->mouse_reporting == MOUSE_SGR; break;
     case 1007: term->xtsave.alt_scrolling = term->alt_scrolling; break;
     case 1015: term->xtsave.mouse_urxvt = term->mouse_reporting == MOUSE_URXVT; break;
+    case 1016: term->xtsave.mouse_sgr_pixels = term->mouse_reporting == MOUSE_SGR_PIXELS; break;
     case 1034: term->xtsave.meta_eight_bit = term->meta.eight_bit; break;
     case 1035: term->xtsave.num_lock_modifier = term->num_lock_modifier; break;
     case 1036: term->xtsave.meta_esc_prefix = term->meta.esc_prefix; break;
     case 1042: term->xtsave.bell_action_enabled = term->bell_action_enabled; break;
+    case 1047: term->xtsave.alt_screen = term->grid == &term->alt; break;
+    case 1048: term_save_cursor(term); break;
     case 1049: term->xtsave.alt_screen = term->grid == &term->alt; break;
+    case 1070: term->xtsave.sixel_private_palette = term->sixel.use_private_palette; break;
     case 2004: term->xtsave.bracketed_paste = term->bracketed_paste; break;
+    case 2026: term->xtsave.app_sync_updates = term->render.app_sync_updates.enabled; break;
+    case 8452: term->xtsave.sixel_cursor_right_of_graphics = term->sixel.cursor_right_of_graphics; break;
     case 27127: term->xtsave.modify_escape_key = term->modify_escape_key; break;
     case 737769: term->xtsave.ime = term_ime_is_enabled(term); break;
     }
@@ -648,6 +693,8 @@ xtrestore(struct terminal *term, unsigned param)
     case 12: enable = term->xtsave.cursor_blink; break;
     case 25: enable = term->xtsave.show_cursor; break;
     case 45: enable = term->xtsave.reverse_wrap; break;
+    case 47: enable = term->xtsave.alt_screen; break;
+    case 80: enable = term->xtsave.sixel_display_mode; break;
     case 1000: enable = term->xtsave.mouse_click; break;
     case 1001: return;
     case 1002: enable = term->xtsave.mouse_drag; break;
@@ -657,12 +704,18 @@ xtrestore(struct terminal *term, unsigned param)
     case 1006: enable = term->xtsave.mouse_sgr; break;
     case 1007: enable = term->xtsave.alt_scrolling; break;
     case 1015: enable = term->xtsave.mouse_urxvt; break;
+    case 1016: enable = term->xtsave.mouse_sgr_pixels; break;
     case 1034: enable = term->xtsave.meta_eight_bit; break;
     case 1035: enable = term->xtsave.num_lock_modifier; break;
     case 1036: enable = term->xtsave.meta_esc_prefix; break;
     case 1042: enable = term->xtsave.bell_action_enabled; break;
+    case 1047: enable = term->xtsave.alt_screen; break;
+    case 1048: enable = true; break;
     case 1049: enable = term->xtsave.alt_screen; break;
+    case 1070: enable = term->xtsave.sixel_private_palette; break;
     case 2004: enable = term->xtsave.bracketed_paste; break;
+    case 2026: enable = term->xtsave.app_sync_updates; break;
+    case 8452: enable = term->xtsave.sixel_cursor_right_of_graphics; break;
     case 27127: enable = term->xtsave.modify_escape_key; break;
     case 737769: enable = term->xtsave.ime; break;
 
@@ -690,7 +743,7 @@ csi_dispatch(struct terminal *term, uint8_t final)
                 int count = vt_param_get(term, 0, 1);
                 LOG_DBG("REP: '%lc' %d times", (wint_t)term->vt.last_printed, count);
 
-                const int width = wcwidth(term->vt.last_printed);
+                const int width = c32width(term->vt.last_printed);
                 if (width > 0) {
                     for (int i = 0; i < count; i++)
                         term_print(term, term->vt.last_printed, width);
@@ -741,8 +794,8 @@ csi_dispatch(struct terminal *term, uint8_t final)
              *
              * Note: tertiary DA responds with "FOOT".
              */
-            const char *reply = "\033[?62;4;22c";
-            term_to_slave(term, reply, strlen(reply));
+            static const char reply[] = "\033[?62;4;22c";
+            term_to_slave(term, reply, sizeof(reply) - 1);
             break;
         }
 
@@ -837,52 +890,34 @@ csi_dispatch(struct terminal *term, uint8_t final)
 
             int param = vt_param_get(term, 0, 0);
             switch (param) {
-            case 0:
+            case 0: {
                 /* From cursor to end of screen */
+                const struct coord *cursor = &term->grid->cursor.point;
                 term_erase(
                     term,
-                    &term->grid->cursor.point,
-                    &(struct coord){term->cols - 1, term->rows - 1});
+                    cursor->row, cursor->col,
+                    term->rows - 1, term->cols - 1);
                 term->grid->cursor.lcf = false;
                 break;
+            }
 
-            case 1:
+            case 1: {
                 /* From start of screen to cursor */
-                term_erase(term, &(struct coord){0, 0}, &term->grid->cursor.point);
+                const struct coord *cursor = &term->grid->cursor.point;
+                term_erase(term, 0, 0, cursor->row, cursor->col);
                 term->grid->cursor.lcf = false;
                 break;
+            }
 
             case 2:
                 /* Erase entire screen */
-                term_erase(
-                    term,
-                    &(struct coord){0, 0},
-                    &(struct coord){term->cols - 1, term->rows - 1});
+                term_erase(term, 0, 0, term->rows - 1, term->cols - 1);
                 term->grid->cursor.lcf = false;
                 break;
 
             case 3: {
                 /* Erase scrollback */
-                int end = (term->grid->offset + term->rows - 1) % term->grid->num_rows;
-                for (size_t i = 0; i < term->grid->num_rows; i++) {
-                    if (end >= term->grid->offset) {
-                        /* Not wrapped */
-                        if (i >= term->grid->offset && i <= end)
-                            continue;
-                    } else {
-                        /* Wrapped */
-                        if (i >= term->grid->offset || i <= end)
-                            continue;
-                    }
-
-                    if (term->render.last_cursor.row == term->grid->rows[i])
-                        term->render.last_cursor.row = NULL;
-
-                    grid_row_free(term->grid->rows[i]);
-                    term->grid->rows[i] = NULL;
-                }
-                term->grid->view = term->grid->offset;
-                term_damage_view(term);
+                term_erase_scrollback(term);
                 break;
             }
 
@@ -898,30 +933,32 @@ csi_dispatch(struct terminal *term, uint8_t final)
 
             int param = vt_param_get(term, 0, 0);
             switch (param) {
-            case 0:
+            case 0: {
                 /* From cursor to end of line */
+                const struct coord *cursor = &term->grid->cursor.point;
                 term_erase(
                     term,
-                    &term->grid->cursor.point,
-                    &(struct coord){term->cols - 1, term->grid->cursor.point.row});
+                    cursor->row, cursor->col,
+                    cursor->row, term->cols - 1);
                 term->grid->cursor.lcf = false;
                 break;
+            }
 
-            case 1:
+            case 1: {
                 /* From start of line to cursor */
-                term_erase(
-                    term, &(struct coord){0, term->grid->cursor.point.row}, &term->grid->cursor.point);
+                const struct coord *cursor = &term->grid->cursor.point;
+                term_erase(term, cursor->row, 0, cursor->row, cursor->col);
                 term->grid->cursor.lcf = false;
                 break;
+            }
 
-            case 2:
+            case 2: {
                 /* Entire line */
-                term_erase(
-                    term,
-                    &(struct coord){0, term->grid->cursor.point.row},
-                    &(struct coord){term->cols - 1, term->grid->cursor.point.row});
+                const struct coord *cursor = &term->grid->cursor.point;
+                term_erase(term, cursor->row, 0, cursor->row, term->cols - 1);
                 term->grid->cursor.lcf = false;
                 break;
+            }
 
             default:
                 UNHANDLED();
@@ -947,6 +984,7 @@ csi_dispatch(struct terminal *term, uint8_t final)
                     .end = term->scroll_region.end},
                 count);
             term->grid->cursor.lcf = false;
+            term->grid->cursor.point.col = 0;
             break;
         }
 
@@ -966,6 +1004,7 @@ csi_dispatch(struct terminal *term, uint8_t final)
                     .end = term->scroll_region.end},
                 count);
             term->grid->cursor.lcf = false;
+            term->grid->cursor.point.col = 0;
             break;
         }
 
@@ -989,10 +1028,11 @@ csi_dispatch(struct terminal *term, uint8_t final)
             term->grid->cur_row->dirty = true;
 
             /* Erase the remainder of the line */
+            const struct coord *cursor = &term->grid->cursor.point;
             term_erase(
                 term,
-                &(struct coord){term->grid->cursor.point.col + remaining, term->grid->cursor.point.row},
-                &(struct coord){term->cols - 1, term->grid->cursor.point.row});
+                cursor->row, cursor->col + remaining,
+                cursor->row, term->cols - 1);
             term->grid->cursor.lcf = false;
             break;
         }
@@ -1016,26 +1056,25 @@ csi_dispatch(struct terminal *term, uint8_t final)
             term->grid->cur_row->dirty = true;
 
             /* Erase (insert space characters) */
+            const struct coord *cursor = &term->grid->cursor.point;
             term_erase(
                 term,
-                &term->grid->cursor.point,
-                &(struct coord){term->grid->cursor.point.col + count - 1, term->grid->cursor.point.row});
+                cursor->row, cursor->col,
+                cursor->row, cursor->col + count - 1);
             term->grid->cursor.lcf = false;
             break;
         }
 
         case 'S': {
-            int amount = min(
-                vt_param_get(term, 0, 1),
-                term->scroll_region.end - term->scroll_region.start);
+            const struct scroll_region *r = &term->scroll_region;
+            int amount = min(vt_param_get(term, 0, 1), r->end - r->start);
             term_scroll(term, amount);
             break;
         }
 
         case 'T': {
-            int amount = min(
-                vt_param_get(term, 0, 1),
-                term->scroll_region.end - term->scroll_region.start);
+            const struct scroll_region *r = &term->scroll_region;
+            int amount = min(vt_param_get(term, 0, 1), r->end - r->start);
             term_scroll_reverse(term, amount);
             break;
         }
@@ -1045,10 +1084,11 @@ csi_dispatch(struct terminal *term, uint8_t final)
             int count = min(
                 vt_param_get(term, 0, 1), term->cols - term->grid->cursor.point.col);
 
+            const struct coord *cursor = &term->grid->cursor.point;
             term_erase(
                 term,
-                &term->grid->cursor.point,
-                &(struct coord){term->grid->cursor.point.col + count - 1, term->grid->cursor.point.row});
+                cursor->row, cursor->col,
+                cursor->row, cursor->col + count - 1);
             term->grid->cursor.lcf = false;
             break;
         }
@@ -1063,8 +1103,11 @@ csi_dispatch(struct terminal *term, uint8_t final)
                         break;
                     }
                 }
-                assert(new_col >= term->grid->cursor.point.col);
+                xassert(new_col >= term->grid->cursor.point.col);
+
+                bool lcf = term->grid->cursor.lcf;
                 term_cursor_right(term, new_col - term->grid->cursor.point.col);
+                term->grid->cursor.lcf = lcf;
             }
             break;
         }
@@ -1079,7 +1122,7 @@ csi_dispatch(struct terminal *term, uint8_t final)
                         break;
                     }
                 }
-                assert(term->grid->cursor.point.col >= new_col);
+                xassert(term->grid->cursor.point.col >= new_col);
                 term_cursor_left(term, term->grid->cursor.point.col - new_col);
             }
             break;
@@ -1093,6 +1136,7 @@ csi_dispatch(struct terminal *term, uint8_t final)
 
             case 4:   /* Insert Mode - IRM */
                 term->insert_mode = true;
+                term_update_ascii_printer(term);
                 break;
 
             case 12:  /* Send/receive Mode - SRM */
@@ -1113,6 +1157,7 @@ csi_dispatch(struct terminal *term, uint8_t final)
             switch (vt_param_get(term, 0, 0)) {
             case 4:   /* Insert Mode - IRM */
                 term->insert_mode = false;
+                term_update_ascii_printer(term);
                 break;
 
             case 2:   /* Keyboard Action Mode - AM */
@@ -1141,15 +1186,11 @@ csi_dispatch(struct terminal *term, uint8_t final)
         }
 
         case 's':
-            term->grid->saved_cursor = term->grid->cursor;
-            term->vt.saved_attrs = term->vt.attrs;
-            term->saved_charsets = term->charsets;
+            term_save_cursor(term);
             break;
 
         case 'u':
             term_restore_cursor(term, &term->grid->saved_cursor);
-            term->vt.attrs = term->vt.saved_attrs;
-            term->charsets = term->saved_charsets;
             break;
 
         case 't': {
@@ -1181,21 +1222,12 @@ csi_dispatch(struct terminal *term, uint8_t final)
                 break;
 
             case 13: { /* report window position */
-
                 /* We don't know our position - always report (0,0) */
-                int x = -1;
-                int y = -1;
-
+                static const char reply[] = "\033[3;0;0t";
                 switch (vt_param_get(term, 1, 0)) {
-                case 0:
-                    /* window position */
-                    x = y = 0;
-                    break;
-
-                case 2:
-                    /* text area position */
-                    x = term->margins.left;
-                    y = term->margins.top;
+                case 0: /* window position */
+                case 2: /* text area position */
+                    term_to_slave(term, reply, sizeof(reply) - 1);
                     break;
 
                 default:
@@ -1203,12 +1235,6 @@ csi_dispatch(struct terminal *term, uint8_t final)
                     break;
                 }
 
-                if (x >= 0 && y >= 0) {
-                    char reply[64];
-                    size_t n = xsnprintf(reply, sizeof(reply), "\033[3;%d;%dt",
-                             x / term->scale, y / term->scale);
-                    term_to_slave(term, reply, n);
-                }
                 break;
             }
 
@@ -1423,6 +1449,16 @@ csi_dispatch(struct terminal *term, uint8_t final)
             break;
         }
 
+        case 'u': {
+            enum kitty_kbd_flags flags =
+                term->grid->kitty_kbd.flags[term->grid->kitty_kbd.idx];
+
+            char reply[8];
+            int chars = snprintf(reply, sizeof(reply), "\033[?%uu", flags);
+            term_to_slave(term, reply, chars);
+            break;
+        }
+
         default:
             UNHANDLED();
             break;
@@ -1488,19 +1524,21 @@ csi_dispatch(struct terminal *term, uint8_t final)
 
                 case 1: /* modifyCursorKeys */
                 case 2: /* modifyFunctionKeys */
-                case 4: /* modifyOtherKeys */
                     /* Ignored, we always report modifiers */
-                    if (value != 2 && value != -1 &&
-                        !(resource == 4 && value == 1))
-                    {
-                        LOG_WARN("unimplemented: %s = %d",
-                                 resource == 1 ? "modifyCursorKeys" :
-                                 resource == 2 ? "modifyFunctionKeys" :
-                                 resource == 4 ? "modifyOtherKeys" : "<invalid>",
-                                 value);
+                    if (value != 2 && value != -1) {
+                        LOG_WARN(
+                            "unimplemented: %s = %d",
+                            resource == 1 ? "modifyCursorKeys" :
+                            resource == 2 ? "modifyFunctionKeys" : "<invalid>",
+                            value);
                     }
                     break;
 
+                case 4: /* modifyOtherKeys */
+                    term->modify_other_keys_2 = value == 2;
+                    LOG_DBG("modifyOtherKeys=%d", value);
+                    break;
+
                 default:
                     LOG_WARN("invalid resource %d in %s",
                              resource, csi_as_string(term, final, -1));
@@ -1509,6 +1547,41 @@ csi_dispatch(struct terminal *term, uint8_t final)
             }
             break; /* final == 'm' */
 
+        case 'u': {
+            int flags = vt_param_get(term, 0, 0) & KITTY_KBD_SUPPORTED;
+
+            struct grid *grid = term->grid;
+            uint8_t idx = grid->kitty_kbd.idx;
+
+            if (idx + 1 >= ALEN(grid->kitty_kbd.flags)) {
+                /* Stack full, evict oldest by wrapping around */
+                idx = 0;
+            } else
+                idx++;
+
+            grid->kitty_kbd.flags[idx] = flags;
+            grid->kitty_kbd.idx = idx;
+
+            LOG_DBG("kitty kbd: pushed new flags: 0x%03x", flags);
+            break;
+        }
+
+        case 'q': {
+            /* XTVERSION */
+            if (vt_param_get(term, 0, 0) != 0) {
+                UNHANDLED();
+                break;
+            }
+
+            char reply[64];
+            size_t n = xsnprintf(
+                reply, sizeof(reply), "\033P>|foot(%u.%u.%u%s%s)\033\\",
+                FOOT_MAJOR, FOOT_MINOR, FOOT_PATCH,
+                FOOT_EXTRA[0] != '\0' ? "-" : "", FOOT_EXTRA);
+            term_to_slave(term, reply, n);
+            break;
+        }
+
         default:
             UNHANDLED();
             break;
@@ -1517,6 +1590,36 @@ csi_dispatch(struct terminal *term, uint8_t final)
         break; /* private[0] == '>' */
     }
 
+    case '<': {
+        switch (final) {
+        case 'u': {
+            int count = vt_param_get(term, 0, 1);
+            LOG_DBG("kitty kbd: popping %d levels of flags", count);
+
+            struct grid *grid = term->grid;
+            uint8_t idx = grid->kitty_kbd.idx;
+
+            for (int i = 0; i < count; i++) {
+                /* Reset flags. This ensures we get flags=0 when
+                 * over-popping */
+                grid->kitty_kbd.flags[idx] = 0;
+
+                if (idx == 0)
+                    idx = ALEN(grid->kitty_kbd.flags) - 1;
+                else
+                    idx--;
+            }
+
+            grid->kitty_kbd.idx = idx;
+
+            LOG_DBG("kitty kbd: flags after pop: 0x%03x",
+                    term->grid->kitty_kbd.flags[idx]);
+            break;
+        }
+        }
+        break; /* private[0] == ‘<’ */
+    }
+
     case ' ': {
         switch (final) {
         case 'q': {
@@ -1540,7 +1643,7 @@ csi_dispatch(struct terminal *term, uint8_t final)
 
             case 5:         /* blinking bar */
             case 6:         /* steady bar */
-                term->cursor_style = CURSOR_BAR;
+                term->cursor_style = CURSOR_BEAM;
                 break;
 
             default:
@@ -1563,15 +1666,12 @@ csi_dispatch(struct terminal *term, uint8_t final)
     }
 
     case '!': {
-        switch (final) {
-        case 'p':
+        if (final == 'p') {
             term_reset(term, false);
             break;
-
-        default:
-            UNHANDLED();
-            break;
         }
+
+        UNHANDLED();
         break; /* private[0] == '!' */
     }
 
@@ -1597,6 +1697,39 @@ csi_dispatch(struct terminal *term, uint8_t final)
             term_to_slave(term, "\033P!|464f4f54\033\\", 14);  /* FOOT */
             break;
 
+        case 'u': {
+            int flag_set = vt_param_get(term, 0, 0) & KITTY_KBD_SUPPORTED;
+            int mode = vt_param_get(term, 1, 1);
+
+            struct grid *grid = term->grid;
+            uint8_t idx = grid->kitty_kbd.idx;
+
+            switch (mode) {
+            case 1:
+                /* set bits are set, unset bits are reset */
+                grid->kitty_kbd.flags[idx] = flag_set;
+                break;
+
+            case 2:
+                /* set bits are set, unset bits are left unchanged */
+                grid->kitty_kbd.flags[idx] |= flag_set;
+                break;
+
+            case 3:
+                /* set bits are reset, unset bits are left unchanged */
+                grid->kitty_kbd.flags[idx] &= ~flag_set;
+                break;
+
+            default:
+                UNHANDLED();
+                break;
+            }
+
+            LOG_DBG("kitty kbd: flags after update: 0x%03x",
+                    grid->kitty_kbd.flags[idx]);
+            break;
+        }
+
         default:
             UNHANDLED();
             break;
