@@ -13,9 +13,9 @@ struct terminal;
 
 void ime_enable(struct seat *seat);
 void ime_disable(struct seat *seat);
-void ime_update_cursor_rect(struct seat *seat, struct terminal *term);
+void ime_update_cursor_rect(struct seat *seat);
 
+void ime_reset_pending_preedit(struct seat *seat);
+void ime_reset_pending_commit(struct seat *seat);
+void ime_reset_pending(struct seat *seat);
 void ime_reset_preedit(struct seat *seat);
-void ime_reset_commit(struct seat *seat);
-void ime_reset(struct seat *seat);
-void ime_send_cursor_rect(struct seat *seat, struct terminal *term);
