@@ -11,6 +11,8 @@ enum modifier {
     MOD_ALT = 0x4,
     MOD_CTRL = 0x8,
     MOD_META = 0x10,
+    MOD_MODIFY_OTHER_KEYS_STATE1 = 0x20,
+    MOD_MODIFY_OTHER_KEYS_STATE2 = 0x40,
 };
 
 struct key_data {
@@ -41,7 +43,8 @@ static const struct key_data key_escape[] = {
 
 static const struct key_data key_return[] = {
     {MOD_SHIFT,                                 CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;2;13~"},
-    {MOD_ALT,                                   CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\r"},
+    {MOD_ALT | MOD_MODIFY_OTHER_KEYS_STATE1,    CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\r"},
+    {MOD_ALT | MOD_MODIFY_OTHER_KEYS_STATE2,    CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;3;13~"},
     {MOD_SHIFT | MOD_ALT,                       CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;4;13~"},
     {MOD_CTRL,                                  CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;5;13~"},
     {MOD_SHIFT | MOD_CTRL,                      CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;6;13~"},
@@ -60,8 +63,10 @@ static const struct key_data key_return[] = {
 
 /* Tab isn't covered by the regular "modifyOtherKeys" handling */
 static const struct key_data key_tab[] = {
-    {MOD_SHIFT,                                 CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[Z"},
-    {MOD_ALT,                                   CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;3;9~"},
+    {MOD_SHIFT | MOD_MODIFY_OTHER_KEYS_STATE1,  CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[Z"},
+    {MOD_SHIFT | MOD_MODIFY_OTHER_KEYS_STATE2,  CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;2;9~"},
+    {MOD_ALT | MOD_MODIFY_OTHER_KEYS_STATE1,    CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\t"},
+    {MOD_ALT | MOD_MODIFY_OTHER_KEYS_STATE2,    CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;3;9~"},
     {MOD_SHIFT | MOD_ALT,                       CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;4;9~"},
     {MOD_CTRL,                                  CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;5;9~"},
     {MOD_SHIFT | MOD_CTRL,                      CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;6;9~"},
@@ -78,23 +83,56 @@ static const struct key_data key_tab[] = {
     {MOD_ANY,                                   CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\t"},
 };
 
+/*
+ * Shift+Tab produces ISO_Left_Tab
+ *
+ * However, all combos (except Shift+Tab) acts as if we pressed
+ * mods+shift+tab.
+ */
+static const struct key_data key_iso_left_tab[] = {
+    {MOD_SHIFT | MOD_ALT,                       CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;4;9~"},
+    {MOD_SHIFT | MOD_CTRL,                      CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;6;9~"},
+    {MOD_SHIFT | MOD_ALT | MOD_CTRL,            CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;8;9~"},
+    {MOD_SHIFT | MOD_META,                      CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;10;9~"},
+    {MOD_SHIFT | MOD_META | MOD_ALT,            CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;12;9~"},
+    {MOD_SHIFT | MOD_META | MOD_CTRL,           CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;14;9~"},
+    {MOD_SHIFT | MOD_META | MOD_ALT | MOD_CTRL, CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;16;9~"},
+    {MOD_ANY,                                   CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[Z"},
+};
+
 static const struct key_data key_backspace[] = {
-    {MOD_SHIFT,                                 CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\x7f"},
-    {MOD_ALT,                                   CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\x7f"},
-    {MOD_SHIFT | MOD_ALT,                       CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\x7f"},
-    {MOD_CTRL,                                  CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\x08"},
-    {MOD_SHIFT | MOD_CTRL,                      CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\x08"},
-    {MOD_ALT | MOD_CTRL,                        CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\x08"},
-    {MOD_SHIFT | MOD_ALT | MOD_CTRL,            CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\x08"},
-    {MOD_META,                                  CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\x7f"},
-    {MOD_META | MOD_SHIFT,                      CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\x7f"},
-    {MOD_META | MOD_ALT,                        CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\x7f"},
-    {MOD_META | MOD_SHIFT | MOD_ALT,            CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\x7f"},
-    {MOD_META | MOD_CTRL,                       CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\x08"},
-    {MOD_META | MOD_SHIFT | MOD_CTRL,           CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\x08"},
-    {MOD_META | MOD_ALT | MOD_CTRL,             CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\x08"},
-    {MOD_META | MOD_SHIFT | MOD_ALT | MOD_CTRL, CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\x08"},
-    {MOD_ANY,                                   CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\x7f"},
+    {MOD_SHIFT | MOD_MODIFY_OTHER_KEYS_STATE1,                                 CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\x7f"},
+    {MOD_ALT | MOD_MODIFY_OTHER_KEYS_STATE1,                                   CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\x7f"},
+    {MOD_SHIFT | MOD_ALT | MOD_MODIFY_OTHER_KEYS_STATE1,                       CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\x7f"},
+    {MOD_SHIFT | MOD_CTRL | MOD_MODIFY_OTHER_KEYS_STATE1,                      CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\x08"},
+    {MOD_ALT | MOD_CTRL | MOD_MODIFY_OTHER_KEYS_STATE1,                        CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\x08"},
+    {MOD_SHIFT | MOD_ALT | MOD_CTRL | MOD_MODIFY_OTHER_KEYS_STATE1,            CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\x08"},
+    {MOD_META | MOD_MODIFY_OTHER_KEYS_STATE1,                                  CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\x7f"},
+    {MOD_META | MOD_SHIFT | MOD_MODIFY_OTHER_KEYS_STATE1,                      CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\x7f"},
+    {MOD_META | MOD_ALT | MOD_MODIFY_OTHER_KEYS_STATE1,                        CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\x7f"},
+    {MOD_META | MOD_SHIFT | MOD_ALT | MOD_MODIFY_OTHER_KEYS_STATE1,            CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\x7f"},
+    {MOD_META | MOD_CTRL | MOD_MODIFY_OTHER_KEYS_STATE1,                       CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\x08"},
+    {MOD_META | MOD_SHIFT | MOD_CTRL | MOD_MODIFY_OTHER_KEYS_STATE1,           CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\x08"},
+    {MOD_META | MOD_ALT | MOD_CTRL | MOD_MODIFY_OTHER_KEYS_STATE1,             CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\x08"},
+    {MOD_META | MOD_SHIFT | MOD_ALT | MOD_CTRL | MOD_MODIFY_OTHER_KEYS_STATE1, CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033\x08"},
+
+    {MOD_SHIFT | MOD_MODIFY_OTHER_KEYS_STATE2,                                 CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;2;127~"},
+    {MOD_ALT | MOD_MODIFY_OTHER_KEYS_STATE2,                                   CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;3;127~"},
+    {MOD_SHIFT | MOD_ALT | MOD_MODIFY_OTHER_KEYS_STATE2,                       CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;4;127~"},
+    {MOD_SHIFT | MOD_CTRL | MOD_MODIFY_OTHER_KEYS_STATE2,                      CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;6;8~"},
+    {MOD_ALT | MOD_CTRL | MOD_MODIFY_OTHER_KEYS_STATE2,                        CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;7;8~"},
+    {MOD_SHIFT | MOD_ALT | MOD_CTRL | MOD_MODIFY_OTHER_KEYS_STATE2,            CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;8;8~"},
+    {MOD_META | MOD_MODIFY_OTHER_KEYS_STATE2,                                  CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;9;127~"},
+    {MOD_META | MOD_SHIFT | MOD_MODIFY_OTHER_KEYS_STATE2,                      CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;10;127~"},
+    {MOD_META | MOD_ALT | MOD_MODIFY_OTHER_KEYS_STATE2,                        CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;11;127~"},
+    {MOD_META | MOD_SHIFT | MOD_ALT | MOD_MODIFY_OTHER_KEYS_STATE2,            CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;12;127~"},
+    {MOD_META | MOD_CTRL | MOD_MODIFY_OTHER_KEYS_STATE2,                       CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;13;8~"},
+    {MOD_META | MOD_SHIFT | MOD_CTRL | MOD_MODIFY_OTHER_KEYS_STATE2,           CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;14;8~"},
+    {MOD_META | MOD_ALT | MOD_CTRL | MOD_MODIFY_OTHER_KEYS_STATE2,             CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;15;8~"},
+    {MOD_META | MOD_SHIFT | MOD_ALT | MOD_CTRL | MOD_MODIFY_OTHER_KEYS_STATE2, CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\033[27;16;8~"},
+
+    {MOD_CTRL,                                                                 CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\x08"},
+    {MOD_ANY,                                                                  CURSOR_KEYS_DONTCARE, KEYPAD_DONTCARE, "\x7f"},
 };
 
 #define DEFAULT_MODS_FOR_SINGLE(sym)                                    \
