@@ -23,6 +23,7 @@
 #include "config.h"
 #include "foot-features.h"
 #include "fdm.h"
+#include "macros.h"
 #include "reaper.h"
 #include "render.h"
 #include "server.h"
@@ -33,6 +34,12 @@
 #include "xmalloc.h"
 #include "xsnprintf.h"
 
+#include "char32.h"
+
+#if !defined(__STDC_UTF_32__) || !__STDC_UTF_32__
+ #error "char32_t does not use UTF-32"
+#endif
+
 static bool
 fdm_sigint(struct fdm *fdm, int signo, void *data)
 {
@@ -44,53 +51,59 @@ static const char *
 version_and_features(void)
 {
     static char buf[256];
-    snprintf(buf, sizeof(buf), "version: %s %cime",
-             FOOT_VERSION, feature_ime() ? '+' : '-');
+    snprintf(buf, sizeof(buf), "version: %s %cpgo %cime %cgraphemes %cassertions",
+             FOOT_VERSION,
+             feature_pgo() ? '+' : '-',
+             feature_ime() ? '+' : '-',
+             feature_graphemes() ? '+' : '-',
+             feature_assertions() ? '+' : '-');
     return buf;
 }
 
 static void
 print_usage(const char *prog_name)
 {
-    printf(
-        "Usage: %s [OPTIONS...]\n"
-        "Usage: %s [OPTIONS...] command [ARGS...]\n"
-        "\n"
-        "Options:\n"
-        "  -c,--config=PATH                      load configuration from PATH ($XDG_CONFIG_HOME/foot/foot.ini)\n"
-        "     --check-config                     verify configuration, exit with 0 if ok, otherwise exit with 1\n"
-        "  -f,--font=FONT                        comma separated list of fonts in fontconfig format (monospace)\n"
-        "  -t,--term=TERM                        value to set the environment variable TERM to (foot)\n"
-        "     --title=TITLE                      initial window title (foot)\n"
-        "  -a,--app-id=ID                        window application ID (foot)\n"
-        "     --maximized                        start in maximized mode\n"
-        "     --fullscreen                       start in fullscreen mode\n"
-        "     --login-shell                      start shell as a login shell\n"
-        "  -w,--window-size-pixels=WIDTHxHEIGHT  initial width and height, in pixels (alternative to '--dimensions')\n"
-        "  -W,--window-size-chars=WIDTHxHEIGHT   initial width and height, in characters (alternative to '--geometry')\n"
-        "  -s,--server[=PATH]                    run as a server (use 'footclient' to start terminals).\n"
-        "                                        Without PATH, $XDG_RUNTIME_DIR/foot-$WAYLAND_DISPLAY.sock will be used.\n"
-        "     --hold                             remain open after child process exits\n"
-        "  -p,--print-pid=FILE|FD                print PID to file or FD (only applicable in server mode)\n"
-        "  -l,--log-colorize=[never|always|auto] enable/disable colorization of log output on stderr\n"
-        "  -s,--log-no-syslog                    disable syslog logging (only applicable in server mode)\n"
-        "  -v,--version                          show the version number and quit\n",
-        prog_name, prog_name);
+    static const char options[] =
+        "\nOptions:\n"
+        "  -c,--config=PATH                         load configuration from PATH ($XDG_CONFIG_HOME/foot/foot.ini)\n"
+        "  -C,--check-config                        verify configuration, exit with 0 if ok, otherwise exit with 1\n"
+        "  -o,--override=[section.]key=value        override configuration option\n"
+        "  -f,--font=FONT                           comma separated list of fonts in fontconfig format (monospace)\n"
+        "  -t,--term=TERM                           value to set the environment variable TERM to (" FOOT_DEFAULT_TERM ")\n"
+        "  -T,--title=TITLE                         initial window title (foot)\n"
+        "  -a,--app-id=ID                           window application ID (foot)\n"
+        "  -m,--maximized                           start in maximized mode\n"
+        "  -F,--fullscreen                          start in fullscreen mode\n"
+        "  -L,--login-shell                         start shell as a login shell\n"
+        "  -D,--working-directory=DIR               directory to start in (CWD)\n"
+        "  -w,--window-size-pixels=WIDTHxHEIGHT     initial width and height, in pixels\n"
+        "  -W,--window-size-chars=WIDTHxHEIGHT      initial width and height, in characters\n"
+        "  -s,--server[=PATH]                       run as a server (use 'footclient' to start terminals).\n"
+        "                                           Without PATH, $XDG_RUNTIME_DIR/foot-$WAYLAND_DISPLAY.sock will be used.\n"
+        "  -H,--hold                                remain open after child process exits\n"
+        "  -p,--print-pid=FILE|FD                   print PID to file or FD (only applicable in server mode)\n"
+        "  -d,--log-level={info|warning|error|none} log level (info)\n"
+        "  -l,--log-colorize=[{never|always|auto}]  enable/disable colorization of log output on stderr\n"
+        "  -s,--log-no-syslog                       disable syslog logging (only applicable in server mode)\n"
+        "  -v,--version                             show the version number and quit\n"
+        "  -e                                       ignored (for compatibility with xterm -e)\n";
+
+    printf("Usage: %s [OPTIONS...]\n", prog_name);
+    printf("Usage: %s [OPTIONS...] command [ARGS...]\n", prog_name);
+    puts(options);
 }
 
 bool
 locale_is_utf8(void)
 {
-    assert(strlen(u8"ö") == 2);
+    static const char u8[] = u8"ö";
+    xassert(strlen(u8) == 2);
 
-    wchar_t w;
-    if (mbtowc(&w, u8"ö", 2) != 2)
+    char32_t w;
+    if (mbrtoc32(&w, u8, 2, &(mbstate_t){0}) != 2)
         return false;
 
-    if (w != U'ö')
-        return false;
-
-    return true;
+    return w == U'ö';
 }
 
 struct shutdown_context {
@@ -143,26 +156,50 @@ print_pid(const char *pid_file, bool *unlink_at_exit)
         return false;
 }
 
+static void
+sanitize_signals(void)
+{
+    sigset_t mask;
+    sigemptyset(&mask);
+    sigprocmask(SIG_SETMASK, &mask, NULL);
+
+    struct sigaction dfl = {.sa_handler = SIG_DFL};
+    sigemptyset(&dfl.sa_mask);
+
+    for (int i = 1; i < SIGRTMAX; i++)
+        sigaction(i, &dfl, NULL);
+}
+
 int
 main(int argc, char *const *argv)
 {
-    int ret = EXIT_FAILURE;
+    /* Custom exit code, to enable users to differentiate between foot
+     * itself failing, and the client application failiing */
+    static const int foot_exit_failure = -26;
+    int ret = foot_exit_failure;
+
+    sanitize_signals();
+
+    /* XDG startup notifications */
+    const char *token = getenv("XDG_ACTIVATION_TOKEN");
+    unsetenv("XDG_ACTIVATION_TOKEN");
 
     /* Startup notifications; we don't support it, but must ensure we
      * don't pass this on to programs launched by us */
     unsetenv("DESKTOP_STARTUP_ID");
 
-    const char *const prog_name = argv[0];
+    const char *const prog_name = argc > 0 ? argv[0] : "<nullptr>";
 
     static const struct option longopts[] =  {
         {"config",                 required_argument, NULL, 'c'},
         {"check-config",           no_argument,       NULL, 'C'},
+        {"override",               required_argument, NULL, 'o'},
         {"term",                   required_argument, NULL, 't'},
         {"title",                  required_argument, NULL, 'T'},
         {"app-id",                 required_argument, NULL, 'a'},
         {"login-shell",            no_argument,       NULL, 'L'},
+        {"working-directory",      required_argument, NULL, 'D'},
         {"font",                   required_argument, NULL, 'f'},
-        {"geometry",               required_argument, NULL, 'g'},  /* Deprecated */
         {"window-size-pixels",     required_argument, NULL, 'w'},
         {"window-size-chars",      required_argument, NULL, 'W'},
         {"server",                 optional_argument, NULL, 's'},
@@ -171,6 +208,7 @@ main(int argc, char *const *argv)
         {"fullscreen",             no_argument,       NULL, 'F'},
         {"presentation-timings",   no_argument,       NULL, 'P'}, /* Undocumented */
         {"print-pid",              required_argument, NULL, 'p'},
+        {"log-level",              required_argument, NULL, 'd'},
         {"log-colorize",           optional_argument, NULL, 'l'},
         {"log-no-syslog",          no_argument,       NULL, 'S'},
         {"version",                no_argument,       NULL, 'v'},
@@ -183,6 +221,7 @@ main(int argc, char *const *argv)
     const char *conf_term = NULL;
     const char *conf_title = NULL;
     const char *conf_app_id = NULL;
+    const char *custom_cwd = NULL;
     bool login_shell = false;
     tll(char *) conf_fonts = tll_init();
     enum conf_size_type conf_size_type = CONF_SIZE_PX;
@@ -196,12 +235,15 @@ main(int argc, char *const *argv)
     bool fullscreen = false;
     bool unlink_pid_file = false;
     const char *pid_file = NULL;
+    enum log_class log_level = LOG_CLASS_INFO;
     enum log_colorize log_colorize = LOG_COLORIZE_AUTO;
     bool log_syslog = true;
     user_notifications_t user_notifications = tll_init();
+    config_override_t overrides = tll_init();
 
     while (true) {
-        int c = getopt_long(argc, argv, "+c:Ct:T:a:Lf:g:w:W:s::HmFPp:l::Svh", longopts, NULL);
+        int c = getopt_long(argc, argv, "+c:Co:t:T:a:LD:f:w:W:s::HmFPp:d:l::Sveh", longopts, NULL);
+
         if (c == -1)
             break;
 
@@ -214,6 +256,10 @@ main(int argc, char *const *argv)
             check_config = true;
             break;
 
+        case 'o':
+            tll_push_back(overrides, optarg);
+            break;
+
         case 't':
             conf_term = optarg;
             break;
@@ -230,6 +276,16 @@ main(int argc, char *const *argv)
             conf_app_id = optarg;
             break;
 
+        case 'D': {
+            struct stat st;
+            if (stat(optarg, &st) < 0 || !(st.st_mode & S_IFDIR)) {
+                fprintf(stderr, "error: %s: not a directory\n", optarg);
+                return ret;
+            }
+            custom_cwd = optarg;
+            break;
+        }
+
         case 'f':
             tll_free_and_free(conf_fonts, free);
             for (char *font = strtok(optarg, ","); font != NULL; font = strtok(NULL, ",")) {
@@ -240,7 +296,7 @@ main(int argc, char *const *argv)
 
                 /* Strip trailing spaces */
                 char *end = font + strlen(font);
-                assert(*end == '\0');
+                xassert(*end == '\0');
                 end--;
                 while (end > font && isspace(*end))
                     *(end--) = '\0';
@@ -252,22 +308,11 @@ main(int argc, char *const *argv)
             }
             break;
 
-        case 'g': {
-            LOG_WARN("deprecated: -g,--geometry command line option. Use -w,--window-size-pixels instead");
-            struct user_notification deprecation = {
-                .kind = USER_NOTIFICATION_DEPRECATED,
-                .text = xstrdup(
-                    "\033[1m-g,--geometry\033[21m command line option. "
-                    "Use \033[1m-w,--window-size-pixels\033[21m instead"),
-            };
-            tll_push_back(user_notifications, deprecation);
-            /* FALLTHROUGH */
-        }
         case 'w': {
             unsigned width, height;
             if (sscanf(optarg, "%ux%u", &width, &height) != 2 || width == 0 || height == 0) {
                 fprintf(stderr, "error: invalid window-size-pixels: %s\n", optarg);
-                return EXIT_FAILURE;
+                return ret;
             }
 
             conf_size_type = CONF_SIZE_PX;
@@ -280,7 +325,7 @@ main(int argc, char *const *argv)
             unsigned width, height;
             if (sscanf(optarg, "%ux%u", &width, &height) != 2 || width == 0 || height == 0) {
                 fprintf(stderr, "error: invalid window-size-chars: %s\n", optarg);
-                return EXIT_FAILURE;
+                return ret;
             }
 
             conf_size_type = CONF_SIZE_CELLS;
@@ -317,6 +362,20 @@ main(int argc, char *const *argv)
             pid_file = optarg;
             break;
 
+        case 'd': {
+            int lvl = log_level_from_string(optarg);
+            if (unlikely(lvl < 0)) {
+                fprintf(
+                    stderr,
+                    "-d,--log-level: %s: argument must be one of %s\n",
+                    optarg,
+                    log_level_string_hint());
+                return ret;
+            }
+            log_level = lvl;
+            break;
+        }
+
         case 'l':
             if (optarg == NULL || strcmp(optarg, "auto") == 0)
                 log_colorize = LOG_COLORIZE_AUTO;
@@ -326,7 +385,7 @@ main(int argc, char *const *argv)
                 log_colorize = LOG_COLORIZE_ALWAYS;
             else {
                 fprintf(stderr, "%s: argument must be one of 'never', 'always' or 'auto'\n", optarg);
-                return EXIT_FAILURE;
+                return ret;
             }
             break;
 
@@ -342,16 +401,21 @@ main(int argc, char *const *argv)
             print_usage(prog_name);
             return EXIT_SUCCESS;
 
+        case 'e':
+            break;
+
         case '?':
-            return EXIT_FAILURE;
+            return ret;
         }
     }
 
     log_init(log_colorize, as_server && log_syslog,
-             as_server ? LOG_FACILITY_DAEMON : LOG_FACILITY_USER, LOG_CLASS_WARNING);
+             as_server ? LOG_FACILITY_DAEMON : LOG_FACILITY_USER, log_level);
 
-    argc -= optind;
-    argv += optind;
+    if (argc > 0) {
+        argc -= optind;
+        argv += optind;
+    }
 
     LOG_INFO("%s", version_and_features());
 
@@ -360,18 +424,67 @@ main(int argc, char *const *argv)
         if (uname(&name) < 0)
             LOG_ERRNO("uname() failed");
         else
-            LOG_INFO("arch: %s/%zu-bit", name.machine, sizeof(void *) * 8);
+            LOG_INFO("arch: %s %s/%zu-bit",
+                     name.sysname, name.machine, sizeof(void *) * 8);
     }
 
-    setlocale(LC_CTYPE, "");
-    LOG_INFO("locale: %s", setlocale(LC_CTYPE, NULL));
-    if (!locale_is_utf8()) {
-        LOG_ERR("locale is not UTF-8");
+    srand(time(NULL));
+
+    const char *locale = setlocale(LC_CTYPE, "");
+    if (locale == NULL) {
+        LOG_ERR("setlocale() failed");
         return ret;
     }
 
+    LOG_INFO("locale: %s", locale);
+
+    bool bad_locale = !locale_is_utf8();
+    if (bad_locale) {
+        static const char fallback_locales[][12] = {
+            "C.UTF-8",
+            "en_US.UTF-8",
+        };
+
+        /*
+         * Try to force an UTF-8 locale. If we succeed, launch the
+         * user’s shell as usual, but add a user-notification saying
+         * the locale has been changed.
+         */
+        for (size_t i = 0; i < ALEN(fallback_locales); i++) {
+            const char *const fallback_locale = fallback_locales[i];
+
+            if (setlocale(LC_CTYPE, fallback_locale) != NULL) {
+                LOG_WARN("locale '%s' is not UTF-8, using '%s' instead",
+                         locale, fallback_locale);
+
+                user_notification_add_fmt(
+                    &user_notifications, USER_NOTIFICATION_WARNING,
+                    "locale '%s' is not UTF-8, using '%s' instead",
+                    locale, fallback_locale);
+
+                bad_locale = false;
+                break;
+            }
+        }
+
+        if (bad_locale) {
+            LOG_ERR("locale '%s' is not UTF-8, "
+                    "and failed to enable a fallback locale", locale);
+
+            user_notification_add_fmt(
+                &user_notifications, USER_NOTIFICATION_ERROR,
+                "locale '%s' is not UTF-8, "
+                "and failed to enable a fallback locale",
+                locale);
+        }
+    }
+
     struct config conf = {NULL};
-    if (!config_load(&conf, conf_path, &user_notifications, check_config)) {
+    bool conf_successful = config_load(
+        &conf, conf_path, &user_notifications, &overrides, check_config);
+
+    tll_free(overrides);
+    if (!conf_successful) {
         config_free(conf);
         return ret;
     }
@@ -381,6 +494,14 @@ main(int argc, char *const *argv)
         return EXIT_SUCCESS;
     }
 
+    _Static_assert((int)LOG_CLASS_ERROR == (int)FCFT_LOG_CLASS_ERROR,
+                   "fcft log level enum offset");
+    _Static_assert((int)LOG_COLORIZE_ALWAYS == (int)FCFT_LOG_COLORIZE_ALWAYS,
+                   "fcft colorize enum mismatch");
+    fcft_init(
+        (enum fcft_log_colorize)log_colorize,
+        as_server && log_syslog,
+        (enum fcft_log_class)log_level);
     fcft_set_scaling_filter(conf.tweak.fcft_filter);
 
     if (conf_term != NULL) {
@@ -398,17 +519,22 @@ main(int argc, char *const *argv)
     if (login_shell)
         conf.login_shell = true;
     if (tll_length(conf_fonts) > 0) {
-        for (size_t i = 0; i < ALEN(conf.fonts); i++) {
-            tll_foreach(conf.fonts[i], it)
-                config_font_destroy(&it->item);
-            tll_free(conf.fonts[i]);
-        }
+        for (size_t i = 0; i < ALEN(conf.fonts); i++)
+            config_font_list_destroy(&conf.fonts[i]);
+
+        struct config_font_list *font_list = &conf.fonts[0];
+        xassert(font_list->count == 0);
+        xassert(font_list->arr == NULL);
+
+        font_list->arr = xmalloc(
+            tll_length(conf_fonts) * sizeof(font_list->arr[0]));
+
         tll_foreach(conf_fonts, it) {
             struct config_font font;
             if (!config_font_parse(it->item, &font)) {
                 LOG_ERR("%s: invalid font specification", it->item);
             } else
-                tll_push_back(conf.fonts[0], font);
+                font_list->arr[font_list->count++] = font;
         }
         tll_free(conf_fonts);
     }
@@ -428,26 +554,42 @@ main(int argc, char *const *argv)
     conf.presentation_timings = presentation_timings;
     conf.hold_at_exit = hold;
 
+    if (conf.tweak.font_monospace_warn && conf.fonts[0].count > 0) {
+        check_if_font_is_monospaced(
+            conf.fonts[0].arr[0].pattern, &conf.notifications);
+    }
+
+
+    if (bad_locale) {
+        static char *const bad_locale_fake_argv[] = {"/bin/sh", "-c", "", NULL};
+        argc = 1;
+        argv = bad_locale_fake_argv;
+        conf.hold_at_exit = true;
+    }
+
     struct fdm *fdm = NULL;
     struct reaper *reaper = NULL;
     struct wayland *wayl = NULL;
     struct renderer *renderer = NULL;
     struct terminal *term = NULL;
     struct server *server = NULL;
-    struct shutdown_context shutdown_ctx = {.term = &term, .exit_code = EXIT_FAILURE};
+    struct shutdown_context shutdown_ctx = {.term = &term, .exit_code = foot_exit_failure};
 
-    char *cwd = NULL;
-    {
+    const char *cwd = custom_cwd;
+    char *_cwd = NULL;
+
+    if (cwd == NULL) {
         errno = 0;
         size_t buf_len = 1024;
         do {
-            cwd = xrealloc(cwd, buf_len);
-            if (getcwd(cwd, buf_len) == NULL && errno != ERANGE) {
+            _cwd = xrealloc(_cwd, buf_len);
+            if (getcwd(_cwd, buf_len) == NULL && errno != ERANGE) {
                 LOG_ERRNO("failed to get current working directory");
                 goto out;
             }
             buf_len *= 2;
         } while (errno == ERANGE);
+        cwd = _cwd;
     }
 
     shm_set_max_pool_size(conf.tweak.max_shm_pool_size);
@@ -465,12 +607,13 @@ main(int argc, char *const *argv)
         goto out;
 
     if (!as_server && (term = term_init(
-                           &conf, fdm, reaper, wayl, "foot", cwd, argc, argv,
+                           &conf, fdm, reaper, wayl, "foot", cwd, token,
+                           argc, argv,
                            &term_shutdown_cb, &shutdown_ctx)) == NULL) {
         goto out;
     }
-    free(cwd);
-    cwd = NULL;
+    free(_cwd);
+    _cwd = NULL;
 
     if (as_server && (server = server_init(&conf, fdm, reaper, wayl)) == NULL)
         goto out;
@@ -482,8 +625,11 @@ main(int argc, char *const *argv)
         goto out;
     }
 
-    if (sigaction(SIGHUP, &(struct sigaction){.sa_handler = SIG_IGN}, NULL) < 0) {
-        LOG_ERRNO("failed to ignore SIGHUP");
+    const struct sigaction sig_ign = {.sa_handler = SIG_IGN};
+    if (sigaction(SIGHUP, &sig_ign, NULL) < 0 ||
+        sigaction(SIGPIPE, &sig_ign, NULL) < 0)
+    {
+        LOG_ERRNO("failed to ignore SIGHUP+SIGPIPE");
         goto out;
     }
 
@@ -500,10 +646,11 @@ main(int argc, char *const *argv)
             break;
     }
 
-    ret = aborted || tll_length(wayl->terms) == 0 ? EXIT_SUCCESS : EXIT_FAILURE;
+    if (aborted || tll_length(wayl->terms) == 0)
+        ret = EXIT_SUCCESS;
 
 out:
-    free(cwd);
+    free(_cwd);
     server_destroy(server);
     term_destroy(term);
 
@@ -521,6 +668,7 @@ out:
         unlink(pid_file);
 
     LOG_INFO("goodbye");
+    fcft_fini();
     log_deinit();
     return ret == EXIT_SUCCESS && !as_server ? shutdown_ctx.exit_code : ret;
 }
