@@ -1,5 +1,6 @@
 #include "osc.h"
 
+#include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
 #include <errno.h>
@@ -10,11 +11,13 @@
 #include "base64.h"
 #include "config.h"
 #include "grid.h"
+#include "macros.h"
 #include "notify.h"
 #include "render.h"
 #include "selection.h"
 #include "terminal.h"
 #include "uri.h"
+#include "util.h"
 #include "vt.h"
 #include "xmalloc.h"
 #include "xsnprintf.h"
@@ -106,7 +109,7 @@ from_clipboard_cb(char *text, size_t size, void *user)
     struct clip_context *ctx = user;
     struct terminal *term = ctx->term;
 
-    assert(ctx->idx >= 0 && ctx->idx <= 2);
+    xassert(ctx->idx >= 0 && ctx->idx <= 2);
 
     const char *t = text;
     size_t left = size;
@@ -115,11 +118,11 @@ from_clipboard_cb(char *text, size_t size, void *user)
         for (size_t i = ctx->idx; i < 3 && left > 0; i++, t++, left--)
             ctx->buf[ctx->idx++] = *t;
 
-        assert(ctx->idx <= 3);
+        xassert(ctx->idx <= 3);
         if (ctx->idx == 3) {
             char *chunk = base64_encode(ctx->buf, 3);
-            assert(chunk != NULL);
-            assert(strlen(chunk) == 4);
+            xassert(chunk != NULL);
+            xassert(strlen(chunk) == 4);
 
             term_to_slave(term, chunk, 4);
             free(chunk);
@@ -131,16 +134,16 @@ from_clipboard_cb(char *text, size_t size, void *user)
     if (left == 0)
         return;
 
-    assert(ctx->idx == 0);
+    xassert(ctx->idx == 0);
 
     int remaining = left % 3;
     for (int i = remaining; i > 0; i--)
         ctx->buf[ctx->idx++] = text[size - i];
-    assert(ctx->idx == remaining);
+    xassert(ctx->idx == remaining);
 
     char *chunk = base64_encode((const uint8_t *)t, left / 3 * 3);
-    assert(chunk != NULL);
-    assert(strlen(chunk) % 4 == 0);
+    xassert(chunk != NULL);
+    xassert(strlen(chunk) % 4 == 0);
     term_to_slave(term, chunk, strlen(chunk));
     free(chunk);
 }
@@ -157,7 +160,11 @@ from_clipboard_done(void *user)
         term_to_slave(term, res, 4);
     }
 
-    term_to_slave(term, "\033\\", 2);
+    if (term->vt.osc.bel)
+        term_to_slave(term, "\a", 1);
+    else
+        term_to_slave(term, "\033\\", 2);
+
     free(ctx);
 }
 
@@ -259,8 +266,29 @@ osc_flash(struct terminal *term)
 }
 
 static bool
-parse_legacy_color(const char *string, uint32_t *color)
+parse_legacy_color(const char *string, uint32_t *color, bool *_have_alpha,
+                   uint16_t *_alpha)
 {
+    bool have_alpha = false;
+    uint16_t alpha = 0xffff;
+
+    if (string[0] == '[') {
+        /* e.g. \E]11;[50]#00ff00 */
+        const char *start = &string[1];
+
+        errno = 0;
+        char *end;
+        unsigned long percent = strtoul(start, &end, 10);
+
+        if (errno != 0 || *end != ']')
+            return false;
+
+        have_alpha = true;
+        alpha = (0xffff * min(percent, 100) + 50) / 100;
+
+        string = end + 1;
+    }
+
     if (string[0] != '#')
         return false;
 
@@ -297,31 +325,51 @@ parse_legacy_color(const char *string, uint32_t *color)
     uint8_t g = 256 * (rgb[1] / 65536.);
     uint8_t b = 256 * (rgb[2] / 65536.);
 
-    LOG_DBG("legacy: %02x%02x%02x", r, g, b);
+    LOG_DBG("legacy: %02x%02x%02x (alpha=%04x)", r, g, b,
+            have_alpha ? alpha : 0xffff);
+
     *color = r << 16 | g << 8 | b;
+
+    if (_have_alpha != NULL)
+        *_have_alpha = have_alpha;
+    if (_alpha != NULL)
+        *_alpha = alpha;
     return true;
 }
 
 static bool
-parse_rgb(const char *string, uint32_t *color)
+parse_rgb(const char *string, uint32_t *color, bool *_have_alpha,
+          uint16_t *_alpha)
 {
     size_t len = strlen(string);
+    bool have_alpha = len >= 4 && strncmp(string, "rgba", 4) == 0;
 
-    /* Verify we have the minimum required length (for "rgb:x/x/x") */
-    if (len < 3 /* 'rgb' */ + 1 /* ':' */ + 2 /* '/' */ + 3 * 1 /* 3 * 'x' */)
-        return false;
+    /* Verify we have the minimum required length (for "") */
+    if (have_alpha) {
+        if (len < STRLEN("rgba:x/x/x/x"))
+            return false;
+    } else {
+        if (len < STRLEN("rgb:x/x/x"))
+            return false;
+    }
 
-    /* Verify prefix is "rgb:" */
-    if (string[0] != 'r' || string[1] != 'g' || string[2] != 'b' || string[3] != ':')
-        return false;
+    /* Verify prefix is “rgb:” or “rgba:” */
+    if (have_alpha) {
+        if (strncmp(string, "rgba:", 5) != 0)
+            return false;
+        string += 5;
+        len -= 5;
+    } else {
+        if (strncmp(string, "rgb:", 4) != 0)
+            return false;
+        string += 4;
+        len -= 4;
+    }
 
-    string += 4;
-    len -= 4;
+    int rgb[4];
+    int digits[4];
 
-    int rgb[3];
-    int digits[3];
-
-    for (size_t i = 0; i < 3; i++) {
+    for (size_t i = 0; i < (have_alpha ? 4 : 3); i++) {
         for (rgb[i] = 0, digits[i] = 0;
              len > 0 && *string != '/';
              len--, string++, digits[i]++)
@@ -336,7 +384,7 @@ parse_rgb(const char *string, uint32_t *color)
                     c >= 'a' && c <= 'f' ? c - 'a' + 10 : c - 'A' + 10;
         }
 
-        if (i >= 2)
+        if (i >= (have_alpha ? 3 : 2))
             break;
 
         if (len == 0 || *string != '/')
@@ -349,7 +397,20 @@ parse_rgb(const char *string, uint32_t *color)
     uint8_t g = 256 * (rgb[1] / (double)(1 << (4 * digits[1])));
     uint8_t b = 256 * (rgb[2] / (double)(1 << (4 * digits[2])));
 
-    LOG_DBG("rgb: %02x%02x%02x", r, g, b);
+    uint16_t alpha = 0xffff;
+    if (have_alpha)
+        alpha = 65536 * (rgb[3] / (double)(1 << (4 * digits[3])));
+
+    if (have_alpha)
+        LOG_DBG("rgba: %02x%02x%02x (alpha=%04x)", r, g, b, alpha);
+    else
+        LOG_DBG("rgb: %02x%02x%02x", r, g, b);
+
+    if (_have_alpha != NULL)
+        *_have_alpha = have_alpha;
+    if (_alpha != NULL)
+        *_alpha = alpha;
+
     *color = r << 16 | g << 8 | b;
     return true;
 }
@@ -376,6 +437,64 @@ osc_set_pwd(struct terminal *term, char *string)
     free(host);
 }
 
+static void
+osc_uri(struct terminal *term, char *string)
+{
+    /*
+     * \E]8;<params>;URI\e\\
+     *
+     * Params are key=value pairs, separated by ‘:’.
+     *
+     * The only defined key (as of 2020-05-31) is ‘id’, which is used
+     * to group split-up URIs:
+     *
+     * ╔═ file1 ════╗
+     * ║          ╔═ file2 ═══╗
+     * ║http://exa║Lorem ipsum║
+     * ║le.com    ║ dolor sit ║
+     * ║          ║amet, conse║
+     * ╚══════════║ctetur adip║
+     *            ╚═══════════╝
+     *
+     * This lets a terminal emulator highlight both parts at the same
+     * time (e.g. when hovering over one of the parts with the mouse).
+     */
+
+    char *params = string;
+    char *params_end = strchr(params, ';');
+    if (params_end == NULL)
+        return;
+
+    *params_end = '\0';
+    const char *uri = params_end + 1;
+    uint64_t id = (uint64_t)rand() << 32 | rand();
+
+    char *ctx = NULL;
+    for (const char *key_value = strtok_r(params, ":", &ctx);
+         key_value != NULL;
+         key_value = strtok_r(NULL, ":", &ctx))
+    {
+        const char *key = key_value;
+        char *operator = strchr(key_value, '=');
+
+        if (operator == NULL)
+            continue;
+        *operator = '\0';
+
+        const char *value = operator + 1;
+
+        if (strcmp(key, "id") == 0)
+            id = sdbm_hash(value);
+    }
+
+    LOG_DBG("OSC-8: URL=%s, id=%" PRIu64, uri, id);
+
+    if (uri[0] == '\0')
+        term_osc8_close(term);
+    else
+        term_osc8_open(term, id, uri);
+}
+
 static void
 osc_notify(struct terminal *term, char *string)
 {
@@ -384,7 +503,7 @@ osc_notify(struct terminal *term, char *string)
      * (https://pub.phyks.me/scripts/urxvt/notify) is very simple:
      *
      * #!/usr/bin/perl
-     * 
+     *
      * sub on_osc_seq_perl {
      *   my ($term, $osc, $resp) = @_;
      *   if ($osc =~ /^notify;(\S+);(.*)$/) {
@@ -405,60 +524,7 @@ osc_notify(struct terminal *term, char *string)
     const char *title = strtok_r(string, ";", &ctx);
     const char *msg = strtok_r(NULL, "\x00", &ctx);
 
-    notify_notify(term, title, msg);
-}
-
-static void
-update_color_in_grids(struct terminal *term, uint32_t old_color,
-                      uint32_t new_color)
-{
-    /*
-     * Update color of already rendered cells.
-     *
-     * Note that we do *not* store the original palette
-     * index. Therefore, the best we can do is compare colors - if
-     * they match, assume "our" palette index was the one used to
-     * render the cell.
-     *
-     * There are a couple of cases where this isn't necessarily true:
-     * - user has configured the 16 base colors with non-unique
-     * colors.  - the client has used 24-bit escapes for colors
-     *
-     * In general though, if the client configures the palette, it is
-     * very likely only using index:ed coloring (i.e. not 24-bit
-     * direct colors), and I hope that it is unusual with palettes
-     * where all the colors aren't unique.
-     *
-     * TODO(?): for performance reasons, we only update the current
-     * screen rows (of both grids). I.e. scrollback is *not* updated.
-     */
-    for (size_t i = 0; i < 2; i++) {
-        struct grid *grid = i == 0 ? &term->normal : &term->alt;
-
-        for (size_t r = 0; r < term->rows; r++) {
-            struct row *row = grid_row(grid, r);
-            assert(row != NULL);
-
-            for (size_t c = 0; c < term->grid->num_cols; c++) {
-                struct cell *cell = &row->cells[c];
-                if (cell->attrs.have_fg &&
-                    cell->attrs.fg == old_color)
-                {
-                    cell->attrs.fg = new_color;
-                    cell->attrs.clean = 0;
-                    row->dirty = true;
-                }
-
-                if ( cell->attrs.have_bg &&
-                    cell->attrs.bg == old_color)
-                {
-                    cell->attrs.bg = new_color;
-                    cell->attrs.clean = 0;
-                    row->dirty = true;
-                }
-            }
-        }
-    }
+    notify_notify(term, title, msg != NULL ? msg : "");
 }
 
 void
@@ -484,7 +550,7 @@ osc_dispatch(struct terminal *term)
         param += c - '0';
     }
 
-    LOG_DBG("OCS: %.*s (param = %d)",
+    LOG_DBG("OSC: %.*s (param = %d)",
             (int)term->vt.osc.idx, term->vt.osc.data, param);
 
     char *string = (char *)&term->vt.osc.data[data_ofs];
@@ -501,7 +567,7 @@ osc_dispatch(struct terminal *term)
         if (*string != ';')
             break;
 
-        assert(*string == ';');
+        xassert(*string == ';');
 
         for (const char *s_idx = strtok(string, ";"), *s_color = strtok(NULL, ";");
              s_idx != NULL && s_color != NULL;
@@ -515,24 +581,31 @@ osc_dispatch(struct terminal *term)
                 idx += c - '0';
             }
 
+            if (idx >= ALEN(term->colors.table)) {
+                LOG_WARN("invalid OSC 4 color index: %u", idx);
+                break;
+            }
+
             /* Client queried for current value */
             if (strlen(s_color) == 1 && s_color[0] == '?') {
                 uint32_t color = term->colors.table[idx];
                 uint8_t r = (color >> 16) & 0xff;
                 uint8_t g = (color >>  8) & 0xff;
                 uint8_t b = (color >>  0) & 0xff;
+                const char *terminator = term->vt.osc.bel ? "\a" : "\033\\";
 
                 char reply[32];
-                size_t n = xsnprintf(reply, sizeof(reply), "\033]4;%u;rgb:%02x/%02x/%02x\033\\",
-                         idx, r, g, b);
+                size_t n = xsnprintf(
+                    reply, sizeof(reply), "\033]4;%u;rgb:%02x/%02x/%02x%s",
+                    idx, r, g, b, terminator);
                 term_to_slave(term, reply, n);
             }
 
             else {
                 uint32_t color;
-                bool color_is_valid = s_color[0] == '#'
-                    ? parse_legacy_color(s_color, &color)
-                    : parse_rgb(s_color, &color);
+                bool color_is_valid = s_color[0] == '#' || s_color[0] == '['
+                    ? parse_legacy_color(s_color, &color, NULL, NULL)
+                    : parse_rgb(s_color, &color, NULL, NULL);
 
                 if (!color_is_valid)
                     continue;
@@ -540,8 +613,8 @@ osc_dispatch(struct terminal *term)
                 LOG_DBG("change color definition for #%u from %06x to %06x",
                         idx, term->colors.table[idx], color);
 
-                update_color_in_grids(term, term->colors.table[idx], color);
                 term->colors.table[idx] = color;
+                term_damage_view(term);
             }
         }
 
@@ -553,9 +626,20 @@ osc_dispatch(struct terminal *term)
         osc_set_pwd(term, string);
         break;
 
+    case 8:
+        osc_uri(term, string);
+        break;
+
+    case 9:
+        /* iTerm2 Growl notifications */
+        osc_notify(term, string);
+        break;
+
     case 10:
-    case 11: {
-        /* Set default foreground/background color */
+    case 11:
+    case 17:
+    case 19: {
+        /* Set default foreground/background/highlight-bg/highlight-fg color */
 
         /* Client queried for current value */
         if (strlen(string) == 1 && string[0] == '?') {
@@ -563,6 +647,7 @@ osc_dispatch(struct terminal *term)
             uint8_t r = (color >> 16) & 0xff;
             uint8_t g = (color >>  8) & 0xff;
             uint8_t b = (color >>  0) & 0xff;
+            const char *terminator = term->vt.osc.bel ? "\a" : "\033\\";
 
             /*
              * Reply in XParseColor format
@@ -570,23 +655,51 @@ osc_dispatch(struct terminal *term)
              */
             char reply[32];
             size_t n = xsnprintf(
-                reply, sizeof(reply), "\033]%u;rgb:%02x/%02x/%02x\033\\",
-                param, r, g, b);
+                reply, sizeof(reply), "\033]%u;rgb:%02x/%02x/%02x%s",
+                param, r, g, b, terminator);
 
             term_to_slave(term, reply, n);
             break;
         }
 
         uint32_t color;
-        if (string[0] == '#' ? !parse_legacy_color(string, &color) : !parse_rgb(string, &color))
+        bool have_alpha = false;
+        uint16_t alpha = 0xffff;
+
+        if (string[0] == '#' || string[0] == '['
+            ? !parse_legacy_color(string, &color, &have_alpha, &alpha)
+            : !parse_rgb(string, &color, &have_alpha, &alpha))
+        {
             break;
+        }
 
         LOG_DBG("change color definition for %s to %06x",
-                param == 10 ? "foreground" : "background", color);
+                param == 10 ? "foreground" :
+                param == 11 ? "background" :
+                param == 17 ? "selection background" :
+                              "selection foreground",
+                color);
 
         switch (param) {
-        case 10: term->colors.fg = color; break;
-        case 11: term->colors.bg = color; break;
+        case 10:
+            term->colors.fg = color;
+            break;
+
+        case 11:
+            term->colors.bg = color;
+            if (have_alpha)
+                term->colors.alpha = alpha;
+            break;
+
+        case 17:
+            term->colors.selection_bg = color;
+            term->colors.use_custom_selection = true;
+            break;
+
+        case 19:
+            term->colors.selection_fg = color;
+            term->colors.use_custom_selection = true;
+            break;
         }
 
         term_damage_view(term);
@@ -601,16 +714,25 @@ osc_dispatch(struct terminal *term)
             uint8_t r = (term->cursor_color.cursor >> 16) & 0xff;
             uint8_t g = (term->cursor_color.cursor >>  8) & 0xff;
             uint8_t b = (term->cursor_color.cursor >>  0) & 0xff;
+            const char *terminator = term->vt.osc.bel ? "\a" : "\033\\";
 
             char reply[32];
-            size_t n = xsnprintf(reply, sizeof(reply), "\033]12;rgb:%02x/%02x/%02x\033\\", r, g, b);
+            size_t n = xsnprintf(
+                reply, sizeof(reply), "\033]12;rgb:%02x/%02x/%02x%s",
+                r, g, b, terminator);
+
             term_to_slave(term, reply, n);
             break;
         }
 
         uint32_t color;
-        if (string[0] == '#' ? !parse_legacy_color(string, &color) : !parse_rgb(string, &color))
+
+        if (string[0] == '#' || string[0] == '['
+            ? !parse_legacy_color(string, &color, NULL, NULL)
+            : !parse_rgb(string, &color, NULL, NULL))
+        {
             break;
+        }
 
         LOG_DBG("change cursor color to %06x", color);
 
@@ -634,11 +756,8 @@ osc_dispatch(struct terminal *term)
 
         if (strlen(string) == 0) {
             LOG_DBG("resetting all colors");
-            for (size_t i = 0; i < 256; i++) {
-                update_color_in_grids(
-                    term, term->colors.table[i], term->colors.default_table[i]);
-                term->colors.table[i] = term->colors.default_table[i];
-        }
+            for (size_t i = 0; i < ALEN(term->colors.table); i++)
+                term->colors.table[i] = term->conf->colors.table[i];
         }
 
         else {
@@ -653,13 +772,18 @@ osc_dispatch(struct terminal *term)
                     idx += c - '0';
                 }
 
+                if (idx >= ALEN(term->colors.table)) {
+                    LOG_WARN("invalid OSC 104 color index: %u", idx);
+                    continue;
+                }
+
                 LOG_DBG("resetting color #%u", idx);
-                update_color_in_grids(
-                    term, term->colors.table[idx], term->colors.default_table[idx]);
-                term->colors.table[idx] = term->colors.default_table[idx];
+                term->colors.table[idx] = term->conf->colors.table[idx];
             }
+
         }
 
+        term_damage_view(term);
         break;
     }
 
@@ -667,14 +791,15 @@ osc_dispatch(struct terminal *term)
         break;
 
     case 110: /* Reset default text foreground color */
-        LOG_DBG("resetting foreground");
-        term->colors.fg = term->colors.default_fg;
+        LOG_DBG("resetting foreground color");
+        term->colors.fg = term->conf->colors.fg;
         term_damage_view(term);
         break;
 
     case 111: /* Reset default text background color */
-        LOG_DBG("resetting background");
-        term->colors.bg = term->colors.default_bg;
+        LOG_DBG("resetting background color");
+        term->colors.bg = term->conf->colors.bg;
+        term->colors.alpha = term->conf->colors.alpha;
         term_damage_view(term);
         term_damage_margins(term);
         break;
@@ -686,6 +811,18 @@ osc_dispatch(struct terminal *term)
         term_damage_cursor(term);
         break;
 
+    case 117:
+        LOG_DBG("resetting selection background color");
+        term->colors.selection_bg = term->conf->colors.selection_bg;
+        term->colors.use_custom_selection = term->conf->colors.use_custom.selection;
+        break;
+
+    case 119:
+        LOG_DBG("resetting selection foreground color");
+        term->colors.selection_fg = term->conf->colors.selection_fg;
+        term->colors.use_custom_selection = term->conf->colors.use_custom.selection;
+        break;
+
     case 555:
         osc_flash(term);
         break;
@@ -725,7 +862,7 @@ osc_ensure_size(struct terminal *term, size_t required_size)
         return true;
 
     size_t new_size = (required_size + 127) / 128 * 128;
-    assert(new_size > 0);
+    xassert(new_size > 0);
 
     uint8_t *new_data = realloc(term->vt.osc.data, new_size);
     if (new_data == NULL) {
