@@ -74,13 +74,20 @@ render_xcursor_set(struct seat *seat, struct terminal *term, const char *xcursor
     return true;
 }
 
+const char *
+xcursor_for_csd_border(struct terminal *term, int x, int y)
+{
+    return XCURSOR_LEFT_PTR;
+}
+
 struct wl_window *
-wayl_win_init(struct terminal *term)
+wayl_win_init(struct terminal *term, const char *token)
 {
     return NULL;
 }
 
 void wayl_win_destroy(struct wl_window *win) {}
+bool wayl_win_set_urgent(struct wl_window *win) { return true; }
 
 bool
 spawn(struct reaper *reaper, const char *cwd, char *const argv[],
@@ -105,7 +112,7 @@ render_worker_thread(void *_ctx)
 }
 
 struct extraction_context *
-extract_begin(enum selection_kind kind)
+extract_begin(enum selection_kind kind, bool strip_trailing_empty)
 {
     return NULL;
 }
@@ -129,6 +136,7 @@ void cmd_scrollback_down(struct terminal *term, int rows) {}
 
 void ime_enable(struct seat *seat) {}
 void ime_disable(struct seat *seat) {}
+void ime_reset_preedit(struct seat *seat) {}
 
 void
 notify_notify(const struct terminal *term, const char *title, const char *body)
@@ -138,6 +146,24 @@ notify_notify(const struct terminal *term, const char *title, const char *body)
 void reaper_add(struct reaper *reaper, pid_t pid, reaper_cb cb, void *cb_data) {}
 void reaper_del(struct reaper *reaper, pid_t pid) {}
 
+void urls_reset(struct terminal *term) {}
+
+void shm_unref(struct buffer *buf) {}
+void shm_chain_free(struct buffer_chain *chain) {}
+
+struct buffer_chain *
+shm_chain_new(struct wl_shm *shm, bool scrollable, size_t pix_instances)
+{
+    return NULL;
+}
+
+
+void search_selection_cancelled(struct terminal *term) {}
+
+void get_current_modifiers(const struct seat *seat,
+                           xkb_mod_mask_t *effective,
+                           xkb_mod_mask_t *consumed, uint32_t key) {}
+
 
 int
 main(int argc, const char *const *argv)
