@@ -4,6 +4,7 @@
 #include "terminal.h"
 #include "fdm.h"
 #include "wayland.h"
+#include "misc.h"
 
 struct renderer;
 struct renderer *render_init(struct fdm *fdm, struct wayland *wayl);
@@ -16,6 +17,7 @@ void render_refresh(struct terminal *term);
 void render_refresh_csd(struct terminal *term);
 void render_refresh_search(struct terminal *term);
 void render_refresh_title(struct terminal *term);
+void render_refresh_urls(struct terminal *term);
 bool render_xcursor_set(struct seat *seat, struct terminal *term, const char *xcursor);
 
 struct render_worker_context {
@@ -23,3 +25,12 @@ struct render_worker_context {
     struct terminal *term;
 };
 int render_worker_thread(void *_ctx);
+
+struct csd_data {
+    int x;
+    int y;
+    int width;
+    int height;
+};
+
+struct csd_data get_csd_data(const struct terminal *term, enum csd_surface surf_idx);
