@@ -6,7 +6,6 @@
 #include <ctype.h>
 #include <unistd.h>
 #include <errno.h>
-#include <assert.h>
 #include <signal.h>
 #include <termios.h>
 
@@ -18,6 +17,8 @@
 #define LOG_ENABLE_DBG 0
 #include "log.h"
 
+#include "debug.h"
+#include "macros.h"
 #include "terminal.h"
 #include "tokenize.h"
 #include "xmalloc.h"
@@ -78,19 +79,19 @@ emit_one_notification(int fd, const struct user_notification *notif)
 
     switch (notif->kind) {
     case USER_NOTIFICATION_DEPRECATED:
-        prefix = "\033[33;1mdeprecated\033[39;21m: ";
+        prefix = "\033[33;1mdeprecated\033[39;22m: ";
         break;
 
     case USER_NOTIFICATION_WARNING:
-        prefix = "\033[33;1mwarning\033[39;21m: ";
+        prefix = "\033[33;1mwarning\033[39;22m: ";
         break;
 
     case USER_NOTIFICATION_ERROR:
-        prefix = "\033[31;1merror\033[39;21m: ";
+        prefix = "\033[31;1merror\033[39;22m: ";
         break;
     }
 
-    assert(prefix != NULL);
+    xassert(prefix != NULL);
 
     if (write(fd, prefix, strlen(prefix)) < 0 ||
         write(fd, notif->text, strlen(notif->text)) < 0 ||
@@ -143,7 +144,7 @@ emit_notifications(int fd, const user_notifications_t *notifications)
         emit_notifications_of_kind(fd, notifications, USER_NOTIFICATION_DEPRECATED);
 }
 
-static void
+static noreturn void
 slave_exec(int ptmx, char *argv[], int err_fd, bool login_shell,
            const user_notifications_t *notifications)
 {
@@ -267,37 +268,39 @@ slave_spawn(int ptmx, int argc, const char *cwd, char *const *argv,
         close(fork_pipe[0]);  /* Close read end */
 
         if (chdir(cwd) < 0) {
-            const int _errno = errno;
-            LOG_ERRNO("failed to change working directory");
-            (void)!write(fork_pipe[1], &_errno, sizeof(_errno));
-            _exit(_errno);
+            const int errno_copy = errno;
+            LOG_ERRNO("failed to change working directory to %s", cwd);
+            (void)!write(fork_pipe[1], &errno_copy, sizeof(errno_copy));
+            _exit(errno_copy);
         }
 
-        /* Restore signals */
+        /* Restore signal mask, and SIG_IGN'd signals */
+        const struct sigaction dfl = {.sa_handler = SIG_DFL};
         sigset_t mask;
         sigemptyset(&mask);
-        const struct sigaction sa = {.sa_handler = SIG_DFL};
-        if (sigaction(SIGINT, &sa, NULL) < 0 ||
-            sigaction(SIGTERM, &sa, NULL) < 0 ||
-            sigaction(SIGHUP, &sa, NULL) < 0 ||
-            sigprocmask(SIG_SETMASK, &mask, NULL) < 0)
+
+        if (sigprocmask(SIG_SETMASK, &mask, NULL) < 0 ||
+            sigaction(SIGHUP, &dfl, NULL) < 0 ||
+            sigaction(SIGPIPE, &dfl, NULL) < 0)
         {
-            const int _errno = errno;
+            const int errno_copy = errno;
             LOG_ERRNO_P(errno, "failed to restore signals");
-            (void)!write(fork_pipe[1], &_errno, sizeof(_errno));
-            _exit(_errno);
+            (void)!write(fork_pipe[1], &errno_copy, sizeof(errno_copy));
+            _exit(errno_copy);
         }
 
         setenv("TERM", term_env, 1);
         setenv("COLORTERM", "truecolor", 1);
 
+#if defined(FOOT_TERMINFO_PATH)
+        setenv("TERMINFO", FOOT_TERMINFO_PATH, 1);
+#endif
+
         char **_shell_argv = NULL;
         char **shell_argv = NULL;
 
         if (argc == 0) {
-            char *shell_copy = xstrdup(conf_shell);
-            if (!tokenize_cmdline(shell_copy, &_shell_argv)) {
-                free(shell_copy);
+            if (!tokenize_cmdline(conf_shell, &_shell_argv)) {
                 (void)!write(fork_pipe[1], &errno, sizeof(errno));
                 _exit(0);
             }
@@ -316,25 +319,25 @@ slave_spawn(int ptmx, int argc, const char *cwd, char *const *argv,
             setenv("SHELL", shell_argv[0], 1);
 
         slave_exec(ptmx, shell_argv, fork_pipe[1], login_shell, notifications);
-        assert(false);
+        BUG("Unexpected return from slave_exec()");
         break;
 
     default: {
         close(fork_pipe[1]); /* Close write end */
         LOG_DBG("slave has PID %d", pid);
 
-        int _errno;
-        static_assert(sizeof(errno) == sizeof(_errno), "errno size mismatch");
+        int errno_copy;
+        static_assert(sizeof(errno) == sizeof(errno_copy), "errno size mismatch");
 
-        ssize_t ret = read(fork_pipe[0], &_errno, sizeof(_errno));
+        ssize_t ret = read(fork_pipe[0], &errno_copy, sizeof(errno_copy));
         close(fork_pipe[0]);
 
         if (ret < 0) {
             LOG_ERRNO("failed to read from pipe");
             return -1;
-        } else if (ret == sizeof(_errno)) {
+        } else if (ret == sizeof(errno_copy)) {
             LOG_ERRNO_P(
-                _errno, "%s: failed to execute",
+                errno_copy, "%s: failed to execute",
                 argc == 0 ? conf_shell : argv[0]);
             return -1;
         } else
