@@ -3,20 +3,33 @@
 #include <stdint.h>
 #include <stdbool.h>
 #include <stddef.h>
-#include <wchar.h>
 
 #include <threads.h>
 #include <semaphore.h>
 
+#if defined(FOOT_GRAPHEME_CLUSTERING)
+ #include <utf8proc.h>
+#endif
+
 #include <tllist.h>
 #include <fcft/fcft.h>
 
 //#include "config.h"
+#include "composed.h"
+#include "debug.h"
 #include "fdm.h"
 #include "macros.h"
 #include "reaper.h"
+#include "shm.h"
 #include "wayland.h"
 
+enum color_source {
+    COLOR_DEFAULT,
+    COLOR_BASE16,
+    COLOR_BASE256,
+    COLOR_RGB,
+};
+
 /*
  *  Note: we want the cells to be as small as possible. Larger cells
  *  means fewer scrollback lines (or performance drops due to cache
@@ -25,31 +38,33 @@
  * Note that the members are laid out optimized for x86
  */
 struct attributes {
-    uint32_t bold:1;
-    uint32_t dim:1;
-    uint32_t italic:1;
-    uint32_t underline:1;
-    uint32_t strikethrough:1;
-    uint32_t blink:1;
-    uint32_t conceal:1;
-    uint32_t reverse:1;
+    bool bold:1;
+    bool dim:1;
+    bool italic:1;
+    bool underline:1;
+    bool strikethrough:1;
+    bool blink:1;
+    bool conceal:1;
+    bool reverse:1;
     uint32_t fg:24;
 
-    uint32_t clean:1;
-    uint32_t have_fg:1;
-    uint32_t have_bg:1;
-    uint32_t selected:2;
-    uint32_t reserved:3;
+    bool clean:1;
+    enum color_source fg_src:2;
+    enum color_source bg_src:2;
+    bool confined:1;
+    bool selected:1;
+    bool url:1;
     uint32_t bg:24;
 };
-static_assert(sizeof(struct attributes) == 8, "bad size");
+static_assert(sizeof(struct attributes) == 8, "VT attribute struct too large");
 
-#define CELL_COMB_CHARS_LO    0x40000000ul
-#define CELL_COMB_CHARS_HI    0x400ffffful
-#define CELL_MULT_COL_SPACER  0x40100000ul
+/* Last valid Unicode code point is 0x0010FFFFul */
+#define CELL_COMB_CHARS_LO          0x00200000ul
+#define CELL_COMB_CHARS_HI          (CELL_COMB_CHARS_LO + 0x3fffffff)
+#define CELL_SPACER                 (CELL_COMB_CHARS_HI + 1)
 
 struct cell {
-    wchar_t wc;
+    char32_t wc;
     struct attributes attrs;
 };
 static_assert(sizeof(struct cell) == 12, "bad size");
@@ -78,16 +93,26 @@ struct damage {
     int lines;
 };
 
-struct composed {
-    wchar_t base;
-    wchar_t combining[5];
-    uint8_t count;
+struct row_uri_range {
+    int start;
+    int end;
+    uint64_t id;
+    char *uri;
+};
+
+struct row_data {
+    struct {
+        struct row_uri_range *v;
+        uint32_t size;
+        uint32_t count;
+    } uri_ranges;
 };
 
 struct row {
     struct cell *cells;
     bool dirty;
     bool linebreak;
+    struct row_data *extra;
 };
 
 struct sixel {
@@ -98,6 +123,20 @@ struct sixel {
     int rows;
     int cols;
     struct coord pos;
+    bool opaque;
+};
+
+enum kitty_kbd_flags {
+    KITTY_KBD_DISAMBIGUATE = 0x01,
+    KITTY_KBD_REPORT_EVENT = 0x02,
+    KITTY_KBD_REPORT_ALTERNATE = 0x04,
+    KITTY_KBD_REPORT_ALL = 0x08,
+    KITTY_KBD_REPORT_ASSOCIATED = 0x10,
+    KITTY_KBD_SUPPORTED = (KITTY_KBD_DISAMBIGUATE |
+                           KITTY_KBD_REPORT_EVENT |
+                           KITTY_KBD_REPORT_ALTERNATE |
+                           KITTY_KBD_REPORT_ALL |
+                           KITTY_KBD_REPORT_ASSOCIATED),
 };
 
 struct grid {
@@ -106,6 +145,14 @@ struct grid {
     int offset;
     int view;
 
+    /*
+     * Note: the cursor (not the *saved* cursor) could most likely be
+     * global state in the term struct.
+     *
+     * However, we have grid specific functions that does not have
+     * access to the owning term struct, but does need access to the
+     * cursor.
+     */
     struct cursor cursor;
     struct cursor saved_cursor;
 
@@ -114,6 +161,12 @@ struct grid {
 
     tll(struct damage) scroll_damage;
     tll(struct sixel) sixel_images;
+
+    struct {
+        enum kitty_kbd_flags flags[8];
+        uint8_t idx;
+    } kitty_kbd;
+
 };
 
 struct vt_subparams {
@@ -128,18 +181,34 @@ struct vt_param {
 
 struct vt {
     int state;  /* enum state */
-    wchar_t last_printed;
-    wchar_t utf8;
+    char32_t last_printed;
+#if defined(FOOT_GRAPHEME_CLUSTERING)
+    utf8proc_int32_t grapheme_state;
+#endif
+    char32_t utf8;
     struct {
         struct vt_param v[16];
         uint8_t idx;
     } params;
+
     uint32_t private; /* LSB=priv0, MSB=priv3 */
+
+    struct attributes attrs;
+    struct attributes saved_attrs;
+
     struct {
         uint8_t *data;
         size_t size;
         size_t idx;
+        bool bel; /* true if OSC string was terminated by BEL */
     } osc;
+
+    /* Start coordinate for current OSC-8 URI */
+    struct {
+        uint64_t id;
+        char *uri;
+    } osc8;
+
     struct {
         uint8_t *data;
         size_t size;
@@ -147,17 +216,17 @@ struct vt {
         void (*put_handler)(struct terminal *term, uint8_t c);
         void (*unhook_handler)(struct terminal *term);
     } dcs;
-    struct attributes attrs;
-    struct attributes saved_attrs;
 };
 
 enum cursor_origin { ORIGIN_ABSOLUTE, ORIGIN_RELATIVE };
 enum cursor_keys { CURSOR_KEYS_DONTCARE, CURSOR_KEYS_NORMAL, CURSOR_KEYS_APPLICATION };
 enum keypad_keys { KEYPAD_DONTCARE, KEYPAD_NUMERICAL, KEYPAD_APPLICATION };
 enum charset { CHARSET_ASCII, CHARSET_GRAPHIC };
+enum charset_designator { G0, G1, G2, G3 };
 
 struct charsets {
-    int selected;
+    enum charset_designator selected;
+    enum charset_designator saved;
     enum charset set[4]; /* G0-G3 */
 };
 
@@ -176,11 +245,18 @@ enum mouse_reporting {
     MOUSE_UTF8,          /* ?1005h */
     MOUSE_SGR,           /* ?1006h */
     MOUSE_URXVT,         /* ?1015h */
+    MOUSE_SGR_PIXELS,    /* ?1016h */
 };
 
-enum cursor_style { CURSOR_BLOCK, CURSOR_UNDERLINE, CURSOR_BAR };
+enum cursor_style { CURSOR_BLOCK, CURSOR_UNDERLINE, CURSOR_BEAM };
 
-enum selection_kind { SELECTION_NONE, SELECTION_NORMAL, SELECTION_BLOCK };
+enum selection_kind {
+    SELECTION_NONE,
+    SELECTION_CHAR_WISE,
+    SELECTION_WORD_WISE,
+    SELECTION_LINE_WISE,
+    SELECTION_BLOCK
+};
 enum selection_direction {SELECTION_UNDIR, SELECTION_LEFT, SELECTION_RIGHT};
 enum selection_scroll_direction {SELECTION_SCROLL_NOT, SELECTION_SCROLL_UP, SELECTION_SCROLL_DOWN};
 
@@ -196,6 +272,7 @@ enum term_surface {
     TERM_SURF_SEARCH,
     TERM_SURF_SCROLLBACK_INDICATOR,
     TERM_SURF_RENDER_TIMER,
+    TERM_SURF_JUMP_LABEL,
     TERM_SURF_TITLE,
     TERM_SURF_BORDER_LEFT,
     TERM_SURF_BORDER_RIGHT,
@@ -208,11 +285,33 @@ enum term_surface {
 
 typedef tll(struct ptmx_buffer) ptmx_buffer_list_t;
 
+enum url_action { URL_ACTION_COPY, URL_ACTION_LAUNCH };
+struct url {
+    uint64_t id;
+    char *url;
+    char32_t *key;
+    struct coord start;
+    struct coord end;
+    enum url_action action;
+    bool url_mode_dont_change_url_attr; /* Entering/exiting URL mode doesn’t touch the cells’ attr.url */
+    bool osc8;
+    bool duplicate;
+};
+typedef tll(struct url) url_list_t;
+
+/* If px != 0 then px is valid, otherwise pt is valid */
+struct pt_or_px {
+    int16_t px;
+    float pt;
+};
+
 struct terminal {
     struct fdm *fdm;
     struct reaper *reaper;
     const struct config *conf;
 
+    void (*ascii_printer)(struct terminal *term, char32_t c);
+
     pid_t slave;
     int ptmx;
 
@@ -237,6 +336,7 @@ struct terminal {
     bool focus_events;
     bool alt_scrolling;
     bool modify_escape_key;
+    bool modify_other_keys_2;  /* True when modifyOtherKeys=2 (i.e. “CSI >4;2m”) */
     enum cursor_origin origin;
     enum cursor_keys cursor_keys_mode;
     enum keypad_keys keypad_keys_mode;
@@ -257,10 +357,34 @@ struct terminal {
 
     struct fcft_font *fonts[4];
     struct config_font *font_sizes[4];
+    struct pt_or_px font_line_height;
     float font_dpi;
-    int font_scale;
+    bool font_is_sized_by_dpi;
+    int16_t font_x_ofs;
+    int16_t font_y_ofs;
     enum fcft_subpixel font_subpixel;
 
+    struct {
+        struct fcft_glyph **box_drawing;
+        struct fcft_glyph **braille;
+        struct fcft_glyph **legacy;
+
+        #define GLYPH_BOX_DRAWING_FIRST 0x2500
+        #define GLYPH_BOX_DRAWING_LAST  0x259F
+        #define GLYPH_BOX_DRAWING_COUNT \
+            (GLYPH_BOX_DRAWING_LAST - GLYPH_BOX_DRAWING_FIRST + 1)
+
+        #define GLYPH_BRAILLE_FIRST 0x2800
+        #define GLYPH_BRAILLE_LAST  0x28FF
+        #define GLYPH_BRAILLE_COUNT \
+            (GLYPH_BRAILLE_LAST - GLYPH_BRAILLE_FIRST + 1)
+
+        #define GLYPH_LEGACY_FIRST 0x1FB00
+        #define GLYPH_LEGACY_LAST  0x1FB9B
+        #define GLYPH_LEGACY_COUNT \
+            (GLYPH_LEGACY_LAST - GLYPH_LEGACY_FIRST + 1)
+    } custom_glyphs;
+
     bool is_sending_paste_data;
     ptmx_buffer_list_t ptmx_buffers;
     ptmx_buffer_list_t ptmx_paste_buffers;
@@ -275,32 +399,39 @@ struct terminal {
 
     /* Saved DECSET modes - we save the SET state */
     struct {
-        uint32_t origin:1;
-        uint32_t application_cursor_keys:1;
-        uint32_t reverse:1;
-        uint32_t show_cursor:1;
-        uint32_t reverse_wrap:1;
-        uint32_t auto_margin:1;
-        uint32_t cursor_blink:1;
-        uint32_t bracketed_paste:1;
-        uint32_t focus_events:1;
-        uint32_t alt_scrolling:1;
-        //uint32_t mouse_x10:1;
-        uint32_t mouse_click:1;
-        uint32_t mouse_drag:1;
-        uint32_t mouse_motion:1;
-        //uint32_t mouse_utf8:1;
-        uint32_t mouse_sgr:1;
-        uint32_t mouse_urxvt:1;
-        uint32_t meta_eight_bit:1;
-        uint32_t meta_esc_prefix:1;
-        uint32_t num_lock_modifier:1;
-        uint32_t bell_action_enabled:1;
-        uint32_t alt_screen:1;
-        uint32_t modify_escape_key:1;
-        uint32_t ime:1;
+        bool origin:1;
+        bool application_cursor_keys:1;
+        bool reverse:1;
+        bool show_cursor:1;
+        bool reverse_wrap:1;
+        bool auto_margin:1;
+        bool cursor_blink:1;
+        bool bracketed_paste:1;
+        bool focus_events:1;
+        bool alt_scrolling:1;
+        //bool mouse_x10:1;
+        bool mouse_click:1;
+        bool mouse_drag:1;
+        bool mouse_motion:1;
+        //bool mouse_utf8:1;
+        bool mouse_sgr:1;
+        bool mouse_urxvt:1;
+        bool mouse_sgr_pixels:1;
+        bool meta_eight_bit:1;
+        bool meta_esc_prefix:1;
+        bool num_lock_modifier:1;
+        bool bell_action_enabled:1;
+        bool alt_screen:1;
+        bool modify_escape_key:1;
+        bool ime:1;
+        bool app_sync_updates:1;
+
+        bool sixel_display_mode:1;
+        bool sixel_private_palette:1;
+        bool sixel_cursor_right_of_graphics:1;
     } xtsave;
 
+    bool window_title_has_been_set;
     char *window_title;
     tll(char *) window_title_stack;
 
@@ -333,10 +464,9 @@ struct terminal {
         uint32_t bg;
         uint32_t table[256];
         uint16_t alpha;
-
-        uint32_t default_fg;
-        uint32_t default_bg;
-        uint32_t default_table[256];
+        uint32_t selection_fg;
+        uint32_t selection_bg;
+        bool use_custom_selection;
     } colors;
 
     enum cursor_style cursor_style;
@@ -357,6 +487,12 @@ struct terminal {
         struct coord start;
         struct coord end;
         bool ongoing;
+        bool spaces_only; /* SELECTION_SEMANTIC_WORD */
+
+        struct {
+            struct coord start;
+            struct coord end;
+        } pivot;
 
         struct {
             int fd;
@@ -367,7 +503,7 @@ struct terminal {
 
     bool is_searching;
     struct {
-        wchar_t *buf;
+        char32_t *buf;
         size_t len;
         size_t sz;
         size_t cursor;
@@ -377,6 +513,11 @@ struct terminal {
         bool view_followed_offset;
         struct coord match;
         size_t match_len;
+
+        struct {
+            char32_t *buf;
+            size_t len;
+        } last;
     } search;
 
     struct wayland *wl;
@@ -386,12 +527,21 @@ struct terminal {
     enum term_surface active_surface;
 
     struct {
+        struct {
+            struct buffer_chain *grid;
+            struct buffer_chain *search;
+            struct buffer_chain *scrollback_indicator;
+            struct buffer_chain *render_timer;
+            struct buffer_chain *url;
+            struct buffer_chain *csd;
+        } chains;
+
         /* Scheduled for rendering, as soon-as-possible */
         struct {
             bool grid;
             bool csd;
             bool search;
-            bool title;
+            bool urls;
         } refresh;
 
         /* Scheduled for rendering, in the next frame callback */
@@ -399,13 +549,19 @@ struct terminal {
             bool grid;
             bool csd;
             bool search;
-            bool title;
+            bool urls;
         } pending;
 
         bool margins;  /* Someone explicitly requested a refresh of the margins */
         bool urgency;  /* Signal 'urgency' (paint borders red) */
 
-        int scrollback_lines; /* Number of scrollback lines, from conf (TODO: move out from render struct?) */
+        struct {
+            struct timespec last_update;
+            bool is_armed;
+            int timer_fd;
+        } title;
+
+        uint32_t scrollback_lines; /* Number of scrollback lines, from conf (TODO: move out from render struct?) */
 
         struct {
             bool enabled;
@@ -414,7 +570,7 @@ struct terminal {
 
         /* Render threads + synchronization primitives */
         struct {
-            size_t count;
+            uint16_t count;
             sem_t start;
             sem_t done;
             mtx_t lock;
@@ -449,49 +605,56 @@ struct terminal {
         } state;
 
         struct coord pos;    /* Current sixel coordinate */
+        int max_non_empty_row_no;
+        size_t row_byte_ofs; /* Byte position into image, for current row */
         int color_idx;       /* Current palette index */
-        int max_col;         /* Largest column index we've seen (aka the image width) */
-        uint32_t *palette;   /* Color palette */
+        uint32_t *private_palette;   /* Private palette, used when private mode 1070 is enabled */
+        uint32_t *shared_palette;    /* Shared palette, used when private mode 1070 is disabled */
+        uint32_t *palette;   /* Points to either private_palette or shared_palette */
+        uint32_t color;
 
         struct {
             uint32_t *data;  /* Raw image data, in ARGB */
             int width;       /* Image width, in pixels */
             int height;      /* Image height, in pixels */
-            bool autosize;
         } image;
 
+        bool scrolling:1;                 /* Private mode 80 */
+        bool use_private_palette:1;       /* Private mode 1070 */
+        bool cursor_right_of_graphics:1;  /* Private mode 8452 */
+
         unsigned params[5];  /* Collected parameters, for RASTER, COLOR_SPEC */
         unsigned param;      /* Currently collecting parameter, for RASTER, COLOR_SPEC and REPEAT */
         unsigned param_idx;  /* Parameters seen */
 
+        bool transparent_bg;
+        uint32_t default_bg;
+
         /* Application configurable */
         unsigned palette_size;  /* Number of colors in palette */
         unsigned max_width;     /* Maximum image width, in pixels */
         unsigned max_height;    /* Maximum image height, in pixels */
     } sixel;
 
-#if defined(FOOT_IME_ENABLED) && FOOT_IME_ENABLED
-    struct {
-        bool enabled;
-        struct {
-            wchar_t *text;
-            struct cell *cells;
-            int count;
+    /* TODO: wrap in a struct */
+    url_list_t urls;
+    char32_t url_keys[5];
+    bool urls_show_uri_on_jump_label;
+    struct grid *url_grid_snapshot;
 
-            struct {
-                bool hidden;
-                int start;  /* Cell index, inclusive */
-                int end;    /* Cell index, exclusive */
-            } cursor;
-        } preedit;
-    } ime;
+#if defined(FOOT_IME_ENABLED) && FOOT_IME_ENABLED
+    bool ime_enabled;
 #endif
 
-    bool is_shutting_down;
-    bool slave_has_been_reaped;
-    int exit_status;
-    void (*shutdown_cb)(void *data, int exit_code);
-    void *shutdown_data;
+    struct {
+        bool in_progress;
+        bool client_has_terminated;
+        int terminate_timeout_fd;
+        int exit_status;
+
+        void (*cb)(void *data, int exit_code);
+        void *cb_data;
+    } shutdown;
 
     char *foot_exe;
     char *cwd;
@@ -514,12 +677,15 @@ struct config;
 struct terminal *term_init(
     const struct config *conf, struct fdm *fdm, struct reaper *reaper,
     struct wayland *wayl, const char *foot_exe, const char *cwd,
-    int argc, char *const *argv,
+    const char *token, int argc, char *const *argv,
     void (*shutdown_cb)(void *data, int exit_code), void *shutdown_data);
 
 bool term_shutdown(struct terminal *term);
 int term_destroy(struct terminal *term);
 
+void term_update_ascii_printer(struct terminal *term);
+void term_single_shift(struct terminal *term, enum charset_designator idx);
+
 void term_reset(struct terminal *term, bool hard);
 bool term_to_slave(struct terminal *term, const void *data, size_t len);
 bool term_paste_data_to_slave(
@@ -528,9 +694,13 @@ bool term_paste_data_to_slave(
 bool term_font_size_increase(struct terminal *term);
 bool term_font_size_decrease(struct terminal *term);
 bool term_font_size_reset(struct terminal *term);
-bool term_font_dpi_changed(struct terminal *term);
+bool term_font_dpi_changed(struct terminal *term, int old_scale);
 void term_font_subpixel_changed(struct terminal *term);
 
+int term_pt_or_px_as_pixels(
+    const struct terminal *term, const struct pt_or_px *pt_or_px);
+
+
 void term_window_configured(struct terminal *term);
 
 void term_damage_rows(struct terminal *term, int start, int end);
@@ -549,7 +719,10 @@ void term_damage_scroll(
     struct scroll_region region, int lines);
 
 void term_erase(
-    struct terminal *term, const struct coord *start, const struct coord *end);
+    struct terminal *term,
+    int start_row, int start_col,
+    int end_row, int end_col);
+void term_erase_scrollback(struct terminal *term);
 
 int term_row_rel_to_abs(const struct terminal *term, int row);
 void term_cursor_home(struct terminal *term);
@@ -560,7 +733,7 @@ void term_cursor_up(struct terminal *term, int count);
 void term_cursor_down(struct terminal *term, int count);
 void term_cursor_blink_update(struct terminal *term);
 
-void term_print(struct terminal *term, wchar_t wc, int width);
+void term_print(struct terminal *term, char32_t wc, int width);
 
 void term_scroll(struct terminal *term, int rows);
 void term_scroll_reverse(struct terminal *term, int rows);
@@ -576,6 +749,7 @@ void term_reverse_index(struct terminal *term);
 
 void term_arm_blink_timer(struct terminal *term);
 
+void term_save_cursor(struct terminal *term);
 void term_restore_cursor(struct terminal *term, const struct cursor *cursor);
 
 void term_visual_focus_in(struct terminal *term);
@@ -584,14 +758,17 @@ void term_kbd_focus_in(struct terminal *term);
 void term_kbd_focus_out(struct terminal *term);
 void term_mouse_down(
     struct terminal *term, int button, int row, int col,
+    int row_pixels, int col_pixels,
     bool shift, bool alt, bool ctrl);
 void term_mouse_up(
     struct terminal *term, int button, int row, int col,
+    int row_pixels, int col_pixels,
     bool shift, bool alt, bool ctrl);
 void term_mouse_motion(
     struct terminal *term, int button, int row, int col,
+    int row_pixels, int col_pixels,
     bool shift, bool alt, bool ctrl);
-bool term_mouse_grabbed(const struct terminal *term, struct seat *seat);
+bool term_mouse_grabbed(const struct terminal *term, const struct seat *seat);
 void term_xcursor_update(struct terminal *term);
 void term_xcursor_update_for_seat(struct terminal *term, struct seat *seat);
 
@@ -614,6 +791,19 @@ bool term_view_to_text(
 bool term_ime_is_enabled(const struct terminal *term);
 void term_ime_enable(struct terminal *term);
 void term_ime_disable(struct terminal *term);
-void term_ime_reset(struct terminal *term);
+bool term_ime_reset(struct terminal *term);
 void term_ime_set_cursor_rect(
     struct terminal *term, int x, int y, int width, int height);
+
+void term_urls_reset(struct terminal *term);
+void term_collect_urls(struct terminal *term);
+
+void term_osc8_open(struct terminal *term, uint64_t id, const char *uri);
+void term_osc8_close(struct terminal *term);
+
+static inline void term_reset_grapheme_state(struct terminal *term)
+{
+#if defined(FOOT_GRAPHEME_CLUSTERING)
+    term->vt.grapheme_state = 0;
+#endif
+}
