@@ -12,13 +12,9 @@
 
 #include <wayland-client.h>
 #include <wayland-cursor.h>
-#include <xdg-shell.h>
 #include <xkbcommon/xkbcommon-compose.h>
 
 #include <tllist.h>
-#include <xdg-output-unstable-v1.h>
-#include <xdg-decoration-unstable-v1.h>
-#include <text-input-unstable-v3.h>
 
 #define LOG_MODULE "wayland"
 #define LOG_ENABLE_DBG 0
@@ -30,45 +26,71 @@
 #include "input.h"
 #include "render.h"
 #include "selection.h"
+#include "shm.h"
 #include "util.h"
 #include "xmalloc.h"
 
+static void
+csd_reload_font(struct wl_window *win, int old_scale)
+{
+    struct terminal *term = win->term;
+    const struct config *conf = term->conf;
+
+    const int scale = term->scale;
+
+    bool enable_csd = win->csd_mode == CSD_YES && !win->is_fullscreen;
+    if (!enable_csd)
+        return;
+    if (win->csd.font != NULL && scale == old_scale)
+        return;
+
+    fcft_destroy(win->csd.font);
+
+    const char *patterns[conf->csd.font.count];
+    for (size_t i = 0; i < conf->csd.font.count; i++)
+        patterns[i] = conf->csd.font.arr[i].pattern;
+
+    char pixelsize[32];
+    snprintf(pixelsize, sizeof(pixelsize),
+             "pixelsize=%u", conf->csd.title_height * scale * 1 / 2);
+
+    LOG_DBG("loading CSD font \"%s:%s\" (old-scale=%d, scale=%d)",
+            patterns[0], pixelsize, old_scale, scale);
+
+    win->csd.font = fcft_from_name(conf->csd.font.count, patterns, pixelsize);
+}
+
 static void
 csd_instantiate(struct wl_window *win)
 {
     struct wayland *wayl = win->term->wl;
-    assert(wayl != NULL);
+    xassert(wayl != NULL);
 
-    for (size_t i = 0; i < ALEN(win->csd.surface); i++) {
-        assert(win->csd.surface[i] == NULL);
-        assert(win->csd.sub_surface[i] == NULL);
-
-        win->csd.surface[i] = wl_compositor_create_surface(wayl->compositor);
-
-        struct wl_surface *parent = i < CSD_SURF_MINIMIZE
-            ? win->surface : win->csd.surface[CSD_SURF_TITLE];
-
-        win->csd.sub_surface[i] = wl_subcompositor_get_subsurface(
-            wayl->sub_compositor, win->csd.surface[i], parent);
-
-        wl_subsurface_set_sync(win->csd.sub_surface[i]);
-        wl_surface_set_user_data(win->csd.surface[i], win);
-        wl_surface_commit(win->csd.surface[i]);
+    for (size_t i = 0; i < CSD_SURF_MINIMIZE; i++) {
+        bool ret = wayl_win_subsurface_new(win, &win->csd.surface[i]);
+        xassert(ret);
     }
+
+    for (size_t i = CSD_SURF_MINIMIZE; i < CSD_SURF_COUNT; i++) {
+        bool ret = wayl_win_subsurface_new_with_custom_parent(
+            win, win->csd.surface[CSD_SURF_TITLE].surf, &win->csd.surface[i]);
+        xassert(ret);
+    }
+
+    csd_reload_font(win, -1);
 }
 
 static void
 csd_destroy(struct wl_window *win)
 {
-    for (size_t i = 0; i < ALEN(win->csd.surface); i++) {
-        if (win->csd.sub_surface[i] != NULL)
-            wl_subsurface_destroy(win->csd.sub_surface[i]);
-        if (win->csd.surface[i] != NULL)
-            wl_surface_destroy(win->csd.surface[i]);
+    struct terminal *term = win->term;
 
-        win->csd.surface[i] = NULL;
-        win->csd.sub_surface[i] = NULL;
-    }
+    fcft_destroy(term->window->csd.font);
+    term->window->csd.font = NULL;
+
+    for (size_t i = 0; i < ALEN(win->csd.surface); i++)
+        wayl_win_subsurface_destroy(&win->csd.surface[i]);
+    shm_purge(term->render.chains.csd);
 }
 
 static void
@@ -132,6 +154,15 @@ seat_add_text_input(struct seat *seat)
 #endif
 }
 
+static void
+key_bindings_destroy(key_binding_list_t *bindings)
+{
+    tll_foreach(*bindings, it) {
+        tll_free(it->item.key_codes);
+        tll_remove(*bindings, it);
+    }
+}
+
 static void
 seat_destroy(struct seat *seat)
 {
@@ -140,13 +171,9 @@ seat_destroy(struct seat *seat)
 
     tll_free(seat->mouse.buttons);
 
-    tll_foreach(seat->kbd.bindings.key, it)
-        tll_free(it->item.bind.key_codes);
-    tll_free(seat->kbd.bindings.key);
-
-    tll_foreach(seat->kbd.bindings.search, it)
-        tll_free(it->item.bind.key_codes);
-    tll_free(seat->kbd.bindings.search);
+    key_bindings_destroy(&seat->kbd.bindings.key);
+    key_bindings_destroy(&seat->kbd.bindings.search);
+    key_bindings_destroy(&seat->kbd.bindings.url);
 
     tll_free(seat->mouse.bindings);
 
@@ -197,7 +224,7 @@ seat_destroy(struct seat *seat)
     if (seat->wl_seat != NULL)
         wl_seat_release(seat->wl_seat);
 
-    ime_reset(seat);
+    ime_reset_pending(seat);
     free(seat->clipboard.text);
     free(seat->primary.text);
     free(seat->name);
@@ -231,7 +258,7 @@ seat_handle_capabilities(void *data, struct wl_seat *wl_seat,
                          enum wl_seat_capability caps)
 {
     struct seat *seat = data;
-    assert(seat->wl_seat == wl_seat);
+    xassert(seat->wl_seat == wl_seat);
 
     LOG_DBG("%s: keyboard=%s, pointer=%s", seat->name,
             (caps & WL_SEAT_CAPABILITY_KEYBOARD) ? "yes" : "no",
@@ -251,7 +278,7 @@ seat_handle_capabilities(void *data, struct wl_seat *wl_seat,
 
     if (caps & WL_SEAT_CAPABILITY_POINTER) {
         if (seat->wl_pointer == NULL) {
-            assert(seat->pointer.surface == NULL);
+            xassert(seat->pointer.surface == NULL);
             seat->pointer.surface = wl_compositor_create_surface(seat->wayl->compositor);
 
             if (seat->pointer.surface == NULL) {
@@ -297,9 +324,12 @@ update_term_for_output_change(struct terminal *term)
     if (tll_length(term->window->on_outputs) == 0)
         return;
 
+    int old_scale = term->scale;
+
     render_resize(term, term->width / term->scale, term->height / term->scale);
-    term_font_dpi_changed(term);
+    term_font_dpi_changed(term, old_scale);
     term_font_subpixel_changed(term);
+    csd_reload_font(term->window, old_scale);
 }
 
 static void
@@ -310,6 +340,11 @@ update_terms_on_monitor(struct monitor *mon)
     tll_foreach(wayl->terms, it) {
         struct terminal *term = it->item;
 
+        if (term->conf->dpi_aware == DPI_AWARE_AUTO) {
+            update_term_for_output_change(term);
+            continue;
+        }
+
         tll_foreach(term->window->on_outputs, it2) {
             if (it2->item == mon) {
                 update_term_for_output_change(term);
@@ -350,13 +385,19 @@ output_update_ppi(struct monitor *mon)
         break;
     }
 
-    mon->ppi.scaled.x = mon->dim.px_scaled.width / x_inches;
-    mon->ppi.scaled.y = mon->dim.px_scaled.height / y_inches;
+    int scaled_width = mon->dim.px_scaled.width;
+    int scaled_height = mon->dim.px_scaled.height;
 
-    float px_diag = sqrt(
-        pow(mon->dim.px_scaled.width, 2) +
-        pow(mon->dim.px_scaled.height, 2));
+    if (scaled_width == 0 && scaled_height == 0 && mon->scale > 0) {
+        /* Estimate scaled width/height if none has been provided */
+        scaled_width = mon->dim.px_real.width / mon->scale;
+        scaled_height = mon->dim.px_real.height / mon->scale;
+    }
 
+    mon->ppi.scaled.x = scaled_width / x_inches;
+    mon->ppi.scaled.y = scaled_height / y_inches;
+
+    float px_diag = sqrt(pow(scaled_width, 2) + pow(scaled_height, 2));
     mon->dpi = px_diag / mon->inch * mon->scale;
 }
 
@@ -367,6 +408,10 @@ output_geometry(void *data, struct wl_output *wl_output, int32_t x, int32_t y,
                 int32_t transform)
 {
     struct monitor *mon = data;
+
+    free(mon->make);
+    free(mon->model);
+
     mon->dim.mm.width = physical_width;
     mon->dim.mm.height = physical_height;
     mon->inch = sqrt(pow(mon->dim.mm.width, 2) + pow(mon->dim.mm.height, 2)) * 0.03937008;
@@ -374,6 +419,7 @@ output_geometry(void *data, struct wl_output *wl_output, int32_t x, int32_t y,
     mon->model = model != NULL ? xstrdup(model) : NULL;
     mon->subpixel = subpixel;
     mon->transform = transform;
+
     output_update_ppi(mon);
 }
 
@@ -406,11 +452,38 @@ output_scale(void *data, struct wl_output *wl_output, int32_t factor)
     output_update_ppi(mon);
 }
 
+#if defined(WL_OUTPUT_NAME_SINCE_VERSION)
+static void
+output_name(void *data, struct wl_output *wl_output, const char *name)
+{
+    struct monitor *mon = data;
+    free(mon->name);
+    mon->name = name != NULL ? xstrdup(name) : NULL;
+}
+#endif
+
+#if defined(WL_OUTPUT_DESCRIPTION_SINCE_VERSION)
+static void
+output_description(void *data, struct wl_output *wl_output,
+                   const char *description)
+{
+    struct monitor *mon = data;
+    free(mon->description);
+    mon->description = description != NULL ? xstrdup(description) : NULL;
+}
+#endif
+
 static const struct wl_output_listener output_listener = {
     .geometry = &output_geometry,
     .mode = &output_mode,
     .done = &output_done,
     .scale = &output_scale,
+#if defined(WL_OUTPUT_NAME_SINCE_VERSION)
+    .name = &output_name,
+#endif
+#if defined(WL_OUTPUT_DESCRIPTION_SINCE_VERSION)
+    .description = &output_description,
+#endif
 };
 
 static void
@@ -442,6 +515,7 @@ xdg_output_handle_name(void *data, struct zxdg_output_v1 *xdg_output,
                        const char *name)
 {
     struct monitor *mon = data;
+    free(mon->name);
     mon->name = name != NULL ? xstrdup(name) : NULL;
 }
 
@@ -450,6 +524,7 @@ xdg_output_handle_description(void *data, struct zxdg_output_v1 *xdg_output,
                               const char *description)
 {
     struct monitor *mon = data;
+    free(mon->description);
     mon->description = description != NULL ? xstrdup(description) : NULL;
 }
 
@@ -535,6 +610,7 @@ xdg_toplevel_configure(void *data, struct xdg_toplevel *xdg_toplevel,
     bool is_activated = false;
     bool is_fullscreen = false;
     bool is_maximized = false;
+    bool is_resizing = false;
     bool is_tiled_top = false;
     bool is_tiled_bottom = false;
     bool is_tiled_left = false;
@@ -566,11 +642,7 @@ xdg_toplevel_configure(void *data, struct xdg_toplevel *xdg_toplevel,
         case XDG_TOPLEVEL_STATE_TILED_RIGHT:  is_tiled_right = true; break;
         case XDG_TOPLEVEL_STATE_TILED_TOP:    is_tiled_top = true; break;
         case XDG_TOPLEVEL_STATE_TILED_BOTTOM: is_tiled_bottom = true; break;
-
-        case XDG_TOPLEVEL_STATE_RESIZING:
-            /* Ignored */
-            /* TODO: throttle? */
-            break;
+        case XDG_TOPLEVEL_STATE_RESIZING:     is_resizing = true; break;
         }
 
 #if defined(LOG_ENABLE_DBG) && LOG_ENABLE_DBG
@@ -603,17 +675,10 @@ xdg_toplevel_configure(void *data, struct xdg_toplevel *xdg_toplevel,
      */
     struct wl_window *win = data;
 
-    if (!is_fullscreen && win->use_csd == CSD_YES && width > 0 && height > 0) {
-        /*
-         * We include the CSD title bar in our window geometry. Thus,
-         * the height we call render_resize() with must be adjusted,
-         * since it expects the size to refer to the main grid only.
-         */
-        height -= win->term->conf->csd.title_height;
-    }
     win->configure.is_activated = is_activated;
     win->configure.is_fullscreen = is_fullscreen;
     win->configure.is_maximized = is_maximized;
+    win->configure.is_resizing = is_resizing;
     win->configure.is_tiled_top = is_tiled_top;
     win->configure.is_tiled_bottom = is_tiled_bottom;
     win->configure.is_tiled_left = is_tiled_left;
@@ -633,7 +698,7 @@ xdg_toplevel_close(void *data, struct xdg_toplevel *xdg_toplevel)
 
 static const struct xdg_toplevel_listener xdg_toplevel_listener = {
     .configure = &xdg_toplevel_configure,
-    .close = &xdg_toplevel_close,
+    /*.close = */&xdg_toplevel_close,  /* epoll-shim defines a macro ‘close’... */
 };
 
 static void
@@ -646,38 +711,55 @@ xdg_surface_configure(void *data, struct xdg_surface *xdg_surface,
     struct terminal *term = win->term;
 
     bool wasnt_configured = !win->is_configured;
+    bool was_resizing = win->is_resizing;
+    bool csd_was_enabled = win->csd_mode == CSD_YES && !win->is_fullscreen;
+    int new_width = win->configure.width;
+    int new_height = win->configure.height;
+
     win->is_configured = true;
     win->is_maximized = win->configure.is_maximized;
+    win->is_fullscreen = win->configure.is_fullscreen;
+    win->is_resizing = win->configure.is_resizing;
     win->is_tiled_top = win->configure.is_tiled_top;
     win->is_tiled_bottom = win->configure.is_tiled_bottom;
     win->is_tiled_left = win->configure.is_tiled_left;
     win->is_tiled_right = win->configure.is_tiled_right;
-    win->is_tiled = win->is_tiled_top || win->is_tiled_bottom || win->is_tiled_left || win->is_tiled_right;
+    win->is_tiled = (win->is_tiled_top ||
+                     win->is_tiled_bottom ||
+                     win->is_tiled_left ||
+                     win->is_tiled_right);
+    win->csd_mode = win->configure.csd_mode;
 
-    if (win->is_fullscreen != win->configure.is_fullscreen && win->use_csd == CSD_YES) {
-        if (win->configure.is_fullscreen)
-            csd_destroy(win);
-        else
-            csd_instantiate(win);
-    }
-    win->is_fullscreen = win->configure.is_fullscreen;
+    bool enable_csd = win->csd_mode == CSD_YES && !win->is_fullscreen;
+    if (!csd_was_enabled && enable_csd)
+        csd_instantiate(win);
+    else if (csd_was_enabled && !enable_csd)
+        csd_destroy(win);
+
+    if (enable_csd && new_width > 0 && new_height > 0)
+        new_height -= win->term->conf->csd.title_height;
 
     xdg_surface_ack_configure(xdg_surface, serial);
 
-    if (term->window->frame_callback != NULL) {
-        /*
-         * Preempt render scheduling.
-         *
-         * Each configure event require a corresponding new
-         * surface+commit. Thus we cannot just schedule a pending
-         * refresh if there’s already a frame being rendered.
-         */
-        wl_callback_destroy(term->window->frame_callback);
-        term->window->frame_callback = NULL;
-    }
-
-    bool resized = render_resize(
-        term, win->configure.width, win->configure.height);
+#if 1
+    /*
+     * TODO: decide if we should do the last “forced” call when ending
+     * an interactive resize.
+     *
+     * Without it, the last TIOCSWINSZ sent to the client will be a
+     * scheduled one. I.e. there will be a small delay after the user
+     * has *stopped* resizing, and the client application receives the
+     * final size.
+     *
+     * Note: if we also disable content centering while resizing, then
+     * the last, forced, resize *is* necessary.
+     */
+    bool resized = was_resizing && !win->is_resizing
+        ? render_resize_force(term, new_width, new_height)
+        : render_resize(term, new_width, new_height);
+#else
+    bool resized = render_resize(term, new_width, new_height);
+#endif
 
     if (win->configure.is_activated)
         term_visual_focus_in(term);
@@ -708,36 +790,22 @@ xdg_toplevel_decoration_configure(void *data,
 {
     struct wl_window *win = data;
 
-    assert(win->term->conf->csd.preferred != CONF_CSD_PREFER_NONE);
+    xassert(win->term->conf->csd.preferred != CONF_CSD_PREFER_NONE);
     switch (mode) {
     case ZXDG_TOPLEVEL_DECORATION_V1_MODE_CLIENT_SIDE:
         LOG_INFO("using CSD decorations");
-        win->use_csd = CSD_YES;
-        csd_instantiate(win);
+        win->configure.csd_mode = CSD_YES;
         break;
 
     case ZXDG_TOPLEVEL_DECORATION_V1_MODE_SERVER_SIDE:
         LOG_INFO("using SSD decorations");
-        win->use_csd = CSD_NO;
-        csd_destroy(win);
+        win->configure.csd_mode = CSD_NO;
         break;
 
     default:
         LOG_ERR("unimplemented: unknown XDG toplevel decoration mode: %u", mode);
         break;
     }
-
-    if (win->is_configured && win->use_csd == CSD_YES) {
-        struct terminal *term = win->term;
-
-        int scale = term->scale;
-        int width = term->width / scale;
-        int height = term->height / scale;
-
-        /* Take CSD title bar into account */
-        height -= term->conf->csd.title_height;
-        render_resize_force(term, width, height);
-    }
 }
 
 static const struct zxdg_toplevel_decoration_v1_listener xdg_toplevel_decoration_listener = {
@@ -891,12 +959,21 @@ handle_global(void *data, struct wl_registry *registry,
         if (!verify_iface_version(interface, version, required))
             return;
 
+#if defined(WL_OUTPUT_NAME_SINCE_VERSION)
+        const uint32_t preferred = WL_OUTPUT_NAME_SINCE_VERSION;
+#elif defined(WL_OUTPUT_RELEASE_SINCE_VERSION)
+        const uint32_t preferred = WL_OUTPUT_RELEASE_SINCE_VERSION;
+#else
+        const uint32_t preferred = required;
+#endif
+
         struct wl_output *output = wl_registry_bind(
-            wayl->registry, name, &wl_output_interface, min(version, 3));
+            wayl->registry, name, &wl_output_interface, min(version, preferred));
 
         tll_push_back(
             wayl->monitors,
             ((struct monitor){.wayl = wayl, .output = output, .wl_name = name,
+             .scale = 1,
              .use_output_release = version >= WL_OUTPUT_RELEASE_SINCE_VERSION}));
 
         struct monitor *mon = &tll_back(wayl->monitors);
@@ -947,6 +1024,17 @@ handle_global(void *data, struct wl_registry *registry,
         }
     }
 
+#if defined(HAVE_XDG_ACTIVATION)
+    else if (strcmp(interface, xdg_activation_v1_interface.name) == 0) {
+        const uint32_t required = 1;
+        if (!verify_iface_version(interface, version, required))
+            return;
+
+        wayl->xdg_activation = wl_registry_bind(
+            wayl->registry, name, &xdg_activation_v1_interface, required);
+    }
+#endif
+
 #if defined(FOOT_IME_ENABLED) && FOOT_IME_ENABLED
     else if (strcmp(interface, zwp_text_input_manager_v3_interface.name) == 0) {
         const uint32_t required = 1;
@@ -1075,13 +1163,25 @@ fdm_wayl(struct fdm *fdm, int fd, int events, void *data)
             return false;
         }
 
-        while (wl_display_prepare_read(wayl->display) != 0)
-            wl_display_dispatch_pending(wayl->display);
+        while (wl_display_prepare_read(wayl->display) != 0) {
+            if (wl_display_dispatch_pending(wayl->display) < 0) {
+                LOG_ERRNO("failed to dispatch pending Wayland events");
+                return false;
+            }
+        }
     }
 
     if (events & EPOLLHUP) {
         LOG_WARN("disconnected from Wayland");
-        wl_display_cancel_read(wayl->display);
+        /*
+         * Do *not* call wl_display_cancel_read() here.
+         *
+         * Doing so causes later calls to wayl_roundtrip() (called
+         * from term_destroy() -> wayl_win_destroy()) to hang
+         * indefinitely.
+         *
+         * https://codeberg.org/dnkl/foot/issues/651
+         */
         return false;
     }
 
@@ -1142,9 +1242,23 @@ wayl_init(const struct config *conf, struct fdm *fdm)
                 "(wl_data_device_manager not implemented by server)");
         goto out;
     }
+    if (tll_length(wayl->seats) == 0) {
+        LOG_ERR("no seats available (wl_seat interface too old?)");
+        goto out;
+    }
     if (wayl->primary_selection_device_manager == NULL)
         LOG_WARN("no primary selection available");
 
+#if defined(HAVE_XDG_ACTIVATION)
+    if (wayl->xdg_activation == NULL && conf->bell.urgent) {
+#else
+    if (conf->bell.urgent) {
+#endif
+        LOG_WARN(
+            "no XDG activation support; "
+            "bell.urgent will fall back to coloring the window margins red");
+    }
+
     if (conf->presentation_timings && wayl->presentation == NULL) {
         LOG_ERR("presentation time interface not implemented by compositor");
         goto out;
@@ -1218,19 +1332,25 @@ wayl_destroy(struct wayland *wayl)
 
     fdm_hook_del(wayl->fdm, &fdm_hook, FDM_HOOK_PRIORITY_LOW);
 
-    tll_foreach(wayl->monitors, it)
+    tll_foreach(wayl->monitors, it) {
         monitor_destroy(&it->item);
-    tll_free(wayl->monitors);
+        tll_remove(wayl->monitors, it);
+    }
 
-    tll_foreach(wayl->seats, it)
+    tll_foreach(wayl->seats, it) {
         seat_destroy(&it->item);
-    tll_free(wayl->seats);
+        tll_remove(wayl->seats, it);
+    }
 
 #if defined(FOOT_IME_ENABLED) && FOOT_IME_ENABLED
     if (wayl->text_input_manager != NULL)
         zwp_text_input_manager_v3_destroy(wayl->text_input_manager);
 #endif
 
+#if defined(HAVE_XDG_ACTIVATION)
+    if (wayl->xdg_activation != NULL)
+        xdg_activation_v1_destroy(wayl->xdg_activation);
+#endif
     if (wayl->xdg_output_manager != NULL)
         zxdg_output_manager_v1_destroy(wayl->xdg_output_manager);
     if (wayl->shell != NULL)
@@ -1262,7 +1382,7 @@ wayl_destroy(struct wayland *wayl)
 }
 
 struct wl_window *
-wayl_win_init(struct terminal *term)
+wayl_win_init(struct terminal *term, const char *token)
 {
     struct wayland *wayl = term->wl;
     const struct config *conf = term->conf;
@@ -1274,8 +1394,9 @@ wayl_win_init(struct terminal *term)
     }
 
     win->term = term;
-    win->use_csd = CSD_UNKNOWN;
+    win->csd_mode = CSD_UNKNOWN;
     win->csd.move_timeout_fd = -1;
+    win->resize_timeout_fd = -1;
 
     win->surface = wl_compositor_create_surface(wayl->compositor);
     if (win->surface == NULL) {
@@ -1306,7 +1427,7 @@ wayl_win_init(struct terminal *term)
 
     if (conf->csd.preferred == CONF_CSD_PREFER_NONE) {
         /* User specifically do *not* want decorations */
-        win->use_csd = CSD_NO;
+        win->csd_mode = CSD_NO;
         LOG_INFO("window decorations disabled by user");
     } else if (wayl->xdg_decoration_manager != NULL) {
         win->xdg_toplevel_decoration = zxdg_decoration_manager_v1_get_toplevel_decoration(
@@ -1325,25 +1446,32 @@ wayl_win_init(struct terminal *term)
             win->xdg_toplevel_decoration, &xdg_toplevel_decoration_listener, win);
     } else {
         /* No decoration manager - thus we *must* draw our own decorations */
-        win->use_csd = CSD_YES;
-        csd_instantiate(win);
+        win->configure.csd_mode = CSD_YES;
         LOG_WARN("no decoration manager available - using CSDs unconditionally");
     }
 
     wl_surface_commit(win->surface);
 
-    if (conf->tweak.render_timer_osd) {
-        win->render_timer_surface = wl_compositor_create_surface(wayl->compositor);
-        if (win->render_timer_surface == NULL) {
+#if defined(HAVE_XDG_ACTIVATION)
+    /* Complete XDG startup notification */
+    if (token)
+        xdg_activation_v1_activate(wayl->xdg_activation, token, win->surface);
+#endif
+
+    switch (conf->tweak.render_timer) {
+    case RENDER_TIMER_OSD:
+    case RENDER_TIMER_BOTH:
+        if (!wayl_win_subsurface_new(win, &win->render_timer)) {
             LOG_ERR("failed to create render timer surface");
             goto out;
         }
-        win->render_timer_sub_surface = wl_subcompositor_get_subsurface(
-            wayl->sub_compositor, win->render_timer_surface, win->surface);
-        wl_subsurface_set_sync(win->render_timer_sub_surface);
-        wl_surface_set_user_data(win->render_timer_surface, win);
-        wl_surface_commit(win->render_timer_surface);
+        break;
+
+    case RENDER_TIMER_NONE:
+    case RENDER_TIMER_LOG:
+        break;
     }
+
     return win;
 
 out:
@@ -1358,6 +1486,8 @@ wayl_win_destroy(struct wl_window *win)
     if (win == NULL)
         return;
 
+    struct terminal *term = win->term;
+
     if (win->csd.move_timeout_fd != -1)
         close(win->csd.move_timeout_fd);
 
@@ -1370,27 +1500,33 @@ wayl_win_destroy(struct wl_window *win)
      * nor mouse focus).
      */
 
-    if (win->render_timer_surface != NULL) {
-        wl_surface_attach(win->render_timer_surface, NULL, 0, 0);
-        wl_surface_commit(win->render_timer_surface);
+    if (win->render_timer.surf != NULL) {
+        wl_surface_attach(win->render_timer.surf, NULL, 0, 0);
+        wl_surface_commit(win->render_timer.surf);
     }
 
-    if (win->scrollback_indicator_surface != NULL) {
-        wl_surface_attach(win->scrollback_indicator_surface, NULL, 0, 0);
-        wl_surface_commit(win->scrollback_indicator_surface);
+    if (win->scrollback_indicator.surf != NULL) {
+        wl_surface_attach(win->scrollback_indicator.surf, NULL, 0, 0);
+        wl_surface_commit(win->scrollback_indicator.surf);
     }
 
     /* Scrollback search */
-    if (win->search_surface != NULL) {
-        wl_surface_attach(win->search_surface, NULL, 0, 0);
-        wl_surface_commit(win->search_surface);
+    if (win->search.surf != NULL) {
+        wl_surface_attach(win->search.surf, NULL, 0, 0);
+        wl_surface_commit(win->search.surf);
+    }
+
+    /* URLs */
+    tll_foreach(win->urls, it) {
+        wl_surface_attach(it->item.surf.surf, NULL, 0, 0);
+        wl_surface_commit(it->item.surf.surf);
     }
 
     /* CSD */
     for (size_t i = 0; i < ALEN(win->csd.surface); i++) {
-        if (win->csd.surface[i] != NULL) {
-            wl_surface_attach(win->csd.surface[i], NULL, 0, 0);
-            wl_surface_commit(win->csd.surface[i]);
+        if (win->csd.surface[i].surf != NULL) {
+            wl_surface_attach(win->csd.surface[i].surf, NULL, 0, 0);
+            wl_surface_commit(win->csd.surface[i].surf);
         }
     }
 
@@ -1403,19 +1539,27 @@ wayl_win_destroy(struct wl_window *win)
 
     tll_free(win->on_outputs);
 
+    tll_foreach(win->urls, it) {
+        wayl_win_subsurface_destroy(&it->item.surf);
+        tll_remove(win->urls, it);
+    }
+
     csd_destroy(win);
-    if (win->render_timer_sub_surface != NULL)
-        wl_subsurface_destroy(win->render_timer_sub_surface);
-    if (win->render_timer_surface != NULL)
-        wl_surface_destroy(win->render_timer_surface);
-    if (win->scrollback_indicator_sub_surface != NULL)
-        wl_subsurface_destroy(win->scrollback_indicator_sub_surface);
-    if (win->scrollback_indicator_surface != NULL)
-        wl_surface_destroy(win->scrollback_indicator_surface);
-    if (win->search_sub_surface != NULL)
-        wl_subsurface_destroy(win->search_sub_surface);
-    if (win->search_surface != NULL)
-        wl_surface_destroy(win->search_surface);
+    wayl_win_subsurface_destroy(&win->search);
+    wayl_win_subsurface_destroy(&win->scrollback_indicator);
+    wayl_win_subsurface_destroy(&win->render_timer);
+
+    shm_purge(term->render.chains.search);
+    shm_purge(term->render.chains.scrollback_indicator);
+    shm_purge(term->render.chains.render_timer);
+    shm_purge(term->render.chains.grid);
+    shm_purge(term->render.chains.url);
+    shm_purge(term->render.chains.csd);
+
+#if defined(HAVE_XDG_ACTIVATION)
+    if (win->xdg_activation_token != NULL)
+        xdg_activation_token_v1_destroy(win->xdg_activation_token);
+#endif
     if (win->frame_callback != NULL)
         wl_callback_destroy(win->frame_callback);
     if (win->xdg_toplevel_decoration != NULL)
@@ -1428,6 +1572,9 @@ wayl_win_destroy(struct wl_window *win)
         wl_surface_destroy(win->surface);
 
     wayl_roundtrip(win->term->wl);
+
+    if (win->resize_timeout_fd >= 0)
+        fdm_del(win->term->wl->fdm, win->resize_timeout_fd);
     free(win);
 }
 
@@ -1440,25 +1587,32 @@ wayl_reload_xcursor_theme(struct seat *seat, int new_scale)
     }
 
     if (seat->pointer.theme != NULL) {
-        assert(seat->pointer.scale != new_scale);
+        xassert(seat->pointer.scale != new_scale);
         wl_cursor_theme_destroy(seat->pointer.theme);
         seat->pointer.theme = NULL;
         seat->pointer.cursor = NULL;
     }
 
-    const char *xcursor_theme = getenv("XCURSOR_THEME");
     int xcursor_size = 24;
 
     {
         const char *env_cursor_size = getenv("XCURSOR_SIZE");
         if (env_cursor_size != NULL) {
-            unsigned size;
-            if (sscanf(env_cursor_size, "%u", &size) == 1)
+            errno = 0;
+            char *end;
+            int size = (int)strtol(env_cursor_size, &end, 10);
+
+            if (errno == 0 && *end == '\0' && size > 0)
                 xcursor_size = size;
+            else
+                LOG_WARN("XCURSOR_SIZE '%s' is invalid, defaulting to 24",
+                         env_cursor_size);
         }
     }
 
-    LOG_INFO("cursor theme: %s, size: %u, scale: %d",
+    const char *xcursor_theme = getenv("XCURSOR_THEME");
+
+    LOG_INFO("cursor theme: %s, size: %d, scale: %d",
              xcursor_theme, xcursor_size, new_scale);
 
     seat->pointer.theme = wl_cursor_theme_load(
@@ -1495,7 +1649,7 @@ wayl_flush(struct wayland *wayl)
 
         /* Socket buffer is full - need to wait for it to become
            writeable again */
-        assert(errno == EAGAIN);
+        xassert(errno == EAGAIN);
 
         while (true) {
             struct pollfd fds[] = {{.fd = wayl->fd, .events = POLLOUT}};
@@ -1513,7 +1667,7 @@ wayl_flush(struct wayland *wayl)
             if (fds[0].revents & POLLHUP)
                 return;
 
-            assert(fds[0].revents & POLLOUT);
+            xassert(fds[0].revents & POLLOUT);
             break;
         }
     }
@@ -1523,13 +1677,124 @@ void
 wayl_roundtrip(struct wayland *wayl)
 {
     wl_display_cancel_read(wayl->display);
-    wl_display_roundtrip(wayl->display);
+    if (wl_display_roundtrip(wayl->display) < 0) {
+        LOG_ERRNO("failed to roundtrip Wayland display");
+        return;
+    }
 
     /* I suspect the roundtrip above clears the pending queue, and
      * that prepare_read() will always succeed in the first call. But,
      * better safe than sorry... */
 
-    while (wl_display_prepare_read(wayl->display) != 0)
-        wl_display_dispatch_pending(wayl->display);
+    while (wl_display_prepare_read(wayl->display) != 0) {
+        if (wl_display_dispatch_pending(wayl->display) < 0) {
+            LOG_ERRNO("failed to dispatch pending Wayland events");
+            return;
+        }
+    }
     wayl_flush(wayl);
 }
+
+#if defined(HAVE_XDG_ACTIVATION)
+static void
+activation_token_done(void *data, struct xdg_activation_token_v1 *xdg_token,
+                      const char *token)
+{
+    struct wl_window *win = data;
+    struct wayland *wayl = win->term->wl;
+
+    LOG_DBG("activation token: %s", token);
+
+    xdg_activation_v1_activate(wayl->xdg_activation, token, win->surface);
+
+    xassert(win->xdg_activation_token == xdg_token);
+    xdg_activation_token_v1_destroy(xdg_token);
+    win->xdg_activation_token = NULL;
+}
+
+static const struct xdg_activation_token_v1_listener activation_token_listener = {
+    .done = &activation_token_done,
+};
+#endif /* HAVE_XDG_ACTIVATION */
+
+bool
+wayl_win_set_urgent(struct wl_window *win)
+{
+#if defined(HAVE_XDG_ACTIVATION)
+    struct wayland *wayl = win->term->wl;
+
+    if (wayl->xdg_activation == NULL)
+        return false;
+
+    if (win->xdg_activation_token != NULL)
+        return true;
+
+    struct xdg_activation_token_v1 *token =
+        xdg_activation_v1_get_activation_token(wayl->xdg_activation);
+
+    if (token == NULL) {
+        LOG_ERR("failed to retrieve XDG activation token");
+        return false;
+    }
+
+    xdg_activation_token_v1_add_listener(token, &activation_token_listener, win);
+    xdg_activation_token_v1_set_surface(token, win->surface);
+    xdg_activation_token_v1_commit(token);
+    win->xdg_activation_token = token;
+    return true;
+#else
+    return false;
+#endif
+}
+
+bool
+wayl_win_subsurface_new_with_custom_parent(
+    struct wl_window *win, struct wl_surface *parent,
+    struct wl_surf_subsurf *surf)
+{
+    struct wayland *wayl = win->term->wl;
+
+    surf->surf = NULL;
+    surf->sub = NULL;
+
+    struct wl_surface *main_surface
+        = wl_compositor_create_surface(wayl->compositor);
+
+    if (main_surface == NULL)
+        return false;
+
+    struct wl_subsurface *sub = wl_subcompositor_get_subsurface(
+        wayl->sub_compositor, main_surface, parent);
+
+    if (sub == NULL) {
+        wl_surface_destroy(main_surface);
+        return false;
+    }
+
+    wl_surface_set_user_data(main_surface, win);
+    wl_subsurface_set_sync(sub);
+
+    surf->surf = main_surface;
+    surf->sub = sub;
+    return true;
+}
+
+bool
+wayl_win_subsurface_new(struct wl_window *win, struct wl_surf_subsurf *surf)
+{
+    return wayl_win_subsurface_new_with_custom_parent(win, win->surface, surf);
+}
+
+void
+wayl_win_subsurface_destroy(struct wl_surf_subsurf *surf)
+{
+    if (surf == NULL)
+        return;
+    if (surf->sub != NULL)
+        wl_subsurface_destroy(surf->sub);
+    if (surf->surf != NULL)
+        wl_surface_destroy(surf->surf);
+
+    surf->surf = NULL;
+    surf->sub = NULL;
+}
