@@ -2,15 +2,25 @@
 
 #include <stdint.h>
 #include <stdbool.h>
-
-#include <sys/time.h>
+#include <time.h>
+#include <uchar.h>
 
 #include <wayland-client.h>
 #include <xkbcommon/xkbcommon.h>
 
-#include <primary-selection-unstable-v1.h>
+/* Wayland protocols */
 #include <presentation-time.h>
+#include <primary-selection-unstable-v1.h>
+#include <text-input-unstable-v3.h>
+#include <xdg-decoration-unstable-v1.h>
+#include <xdg-output-unstable-v1.h>
+#include <xdg-shell.h>
 
+#if defined(HAVE_XDG_ACTIVATION)
+ #include <xdg-activation-v1.h>
+#endif
+
+#include <fcft/fcft.h>
 #include <tllist.h>
 
 #include "fdm.h"
@@ -18,17 +28,9 @@
 /* Forward declarations */
 struct terminal;
 
-typedef tll(xkb_keycode_t) xkb_keycode_list_t;
-
-struct key_binding {
-    xkb_mod_mask_t mods;
-    xkb_keysym_t sym;
-    xkb_keycode_list_t key_codes;
-};
-typedef tll(struct key_binding) key_binding_list_t;
-
 enum bind_action_normal {
     BIND_ACTION_NONE,
+    BIND_ACTION_NOOP,
     BIND_ACTION_SCROLLBACK_UP_PAGE,
     BIND_ACTION_SCROLLBACK_UP_HALF_PAGE,
     BIND_ACTION_SCROLLBACK_UP_LINE,
@@ -49,34 +51,22 @@ enum bind_action_normal {
     BIND_ACTION_PIPE_SCROLLBACK,
     BIND_ACTION_PIPE_VIEW,
     BIND_ACTION_PIPE_SELECTED,
+    BIND_ACTION_SHOW_URLS_COPY,
+    BIND_ACTION_SHOW_URLS_LAUNCH,
 
     /* Mouse specific actions - i.e. they require a mouse coordinate */
     BIND_ACTION_SELECT_BEGIN,
     BIND_ACTION_SELECT_BEGIN_BLOCK,
     BIND_ACTION_SELECT_EXTEND,
+    BIND_ACTION_SELECT_EXTEND_CHAR_WISE,
     BIND_ACTION_SELECT_WORD,
     BIND_ACTION_SELECT_WORD_WS,
     BIND_ACTION_SELECT_ROW,
 
-    BIND_ACTION_KEY_COUNT = BIND_ACTION_PIPE_SELECTED + 1,
+    BIND_ACTION_KEY_COUNT = BIND_ACTION_SHOW_URLS_LAUNCH + 1,
     BIND_ACTION_COUNT = BIND_ACTION_SELECT_ROW + 1,
 };
 
-struct key_binding_normal {
-    struct key_binding bind;
-    enum bind_action_normal action;
-    char **pipe_argv;
-};
-
-struct mouse_binding {
-    enum bind_action_normal action;
-    xkb_mod_mask_t mods;
-    uint32_t button;
-    int count;
-    char **pipe_argv;
-};
-typedef tll(struct mouse_binding) mouse_binding_list_t;
-
 enum bind_action_search {
     BIND_ACTION_SEARCH_NONE,
     BIND_ACTION_SEARCH_CANCEL,
@@ -100,17 +90,44 @@ enum bind_action_search {
     BIND_ACTION_SEARCH_COUNT,
 };
 
-struct key_binding_search {
-    struct key_binding bind;
-    enum bind_action_search action;
+enum bind_action_url {
+    BIND_ACTION_URL_NONE,
+    BIND_ACTION_URL_CANCEL,
+    BIND_ACTION_URL_TOGGLE_URL_ON_JUMP_LABEL,
+    BIND_ACTION_URL_COUNT,
 };
 
+typedef tll(xkb_keycode_t) xkb_keycode_list_t;
+
+struct key_binding {
+    xkb_mod_mask_t mods;
+    xkb_keysym_t sym;
+    xkb_keycode_list_t key_codes;
+
+    int action; /* enum bind_action_* */
+    char **pipe_argv;
+};
+typedef tll(struct key_binding) key_binding_list_t;
+
+struct mouse_binding {
+    enum bind_action_normal action;
+    xkb_mod_mask_t mods;
+    uint32_t button;
+    int count;
+    char **pipe_argv;
+};
+typedef tll(struct mouse_binding) mouse_binding_list_t;
+
 /* Mime-types we support when dealing with data offers (e.g. copy-paste, or DnD) */
 enum data_offer_mime_type {
     DATA_OFFER_MIME_UNSET,
     DATA_OFFER_MIME_TEXT_PLAIN,
     DATA_OFFER_MIME_TEXT_UTF8,
     DATA_OFFER_MIME_URI_LIST,
+
+    DATA_OFFER_MIME_TEXT_TEXT,
+    DATA_OFFER_MIME_TEXT_STRING,
+    DATA_OFFER_MIME_TEXT_UTF8_STRING,
 };
 
 struct wl_window;
@@ -177,7 +194,14 @@ struct seat {
         xkb_mod_index_t mod_shift;
         xkb_mod_index_t mod_alt;
         xkb_mod_index_t mod_ctrl;
-        xkb_mod_index_t mod_meta;
+        xkb_mod_index_t mod_super;
+        xkb_mod_index_t mod_caps;
+        xkb_mod_index_t mod_num;
+
+        xkb_mod_mask_t bind_significant;
+        xkb_mod_mask_t kitty_significant;
+
+        xkb_mod_mask_t selection_override_modmask;
 
         xkb_keycode_t key_arrow_up;
         xkb_keycode_t key_arrow_down;
@@ -186,11 +210,12 @@ struct seat {
         bool shift;
         bool alt;
         bool ctrl;
-        bool meta;
+        bool super;
 
         struct {
-            tll(struct key_binding_normal) key;
-            tll(struct key_binding_search) search;
+            key_binding_list_t key;
+            key_binding_list_t search;
+            key_binding_list_t url;
         } bindings;
     } kbd;
 
@@ -222,10 +247,10 @@ struct seat {
         /* Double- and triple click state */
         int count;
         int last_released_button;
-        struct timeval last_time;
+        struct timespec last_time;
 
         /* We used a discrete axis event in the current pointer frame */
-        double axis_aggregated;
+        double aggregated[2];
         bool have_discrete;
 
         mouse_binding_list_t bindings;
@@ -253,6 +278,15 @@ struct seat {
                 int32_t cursor_begin;
                 int32_t cursor_end;
             } pending;
+
+            char32_t *text;
+            struct cell *cells;
+            int count;
+            struct {
+                bool hidden;
+                int start;  /* Cell index, inclusive */
+                int end;    /* Cell index, exclusive */
+            } cursor;
         } preedit;
 
         struct  {
@@ -268,6 +302,7 @@ struct seat {
             } pending;
         } surrounding;
 
+        bool focused;
         uint32_t serial;
     } ime;
 #endif
@@ -348,57 +383,72 @@ struct monitor {
     bool use_output_release;
 };
 
+struct wl_surf_subsurf {
+    struct wl_surface *surf;
+    struct wl_subsurface *sub;
+};
+
+struct wl_url {
+    const struct url *url;
+    struct wl_surf_subsurf surf;
+};
+
+enum csd_mode {CSD_UNKNOWN, CSD_NO, CSD_YES};
+
 struct wayland;
 struct wl_window {
     struct terminal *term;
     struct wl_surface *surface;
     struct xdg_surface *xdg_surface;
     struct xdg_toplevel *xdg_toplevel;
+#if defined(HAVE_XDG_ACTIVATION)
+    struct xdg_activation_token_v1 *xdg_activation_token;
+#endif
 
     struct zxdg_toplevel_decoration_v1 *xdg_toplevel_decoration;
 
-    enum {CSD_UNKNOWN, CSD_NO, CSD_YES } use_csd;
+    enum csd_mode csd_mode;
 
     struct {
-        struct wl_surface *surface[CSD_SURF_COUNT];
-        struct wl_subsurface *sub_surface[CSD_SURF_COUNT];
+        struct wl_surf_subsurf surface[CSD_SURF_COUNT];
+        struct fcft_font *font;
         int move_timeout_fd;
         uint32_t serial;
     } csd;
 
-    /* Scrollback search */
-    struct wl_surface *search_surface;
-    struct wl_subsurface *search_sub_surface;
-
-    struct wl_surface *scrollback_indicator_surface;
-    struct wl_subsurface *scrollback_indicator_sub_surface;
-
-    struct wl_surface *render_timer_surface;
-    struct wl_subsurface *render_timer_sub_surface;
+    struct wl_surf_subsurf search;
+    struct wl_surf_subsurf scrollback_indicator;
+    struct wl_surf_subsurf render_timer;
 
     struct wl_callback *frame_callback;
 
     tll(const struct monitor *) on_outputs; /* Outputs we're mapped on */
+    tll(struct wl_url) urls;
 
     bool is_configured;
     bool is_fullscreen;
     bool is_maximized;
+    bool is_resizing;
     bool is_tiled_top;
     bool is_tiled_bottom;
     bool is_tiled_left;
     bool is_tiled_right;
     bool is_tiled;  /* At least one of is_tiled_{top,bottom,left,right} is true */
     struct {
-        bool is_activated;
-        bool is_fullscreen;
-        bool is_maximized;
-        bool is_tiled_top;
-        bool is_tiled_bottom;
-        bool is_tiled_left;
-        bool is_tiled_right;
         int width;
         int height;
+        bool is_activated:1;
+        bool is_fullscreen:1;
+        bool is_maximized:1;
+        bool is_resizing:1;
+        bool is_tiled_top:1;
+        bool is_tiled_bottom:1;
+        bool is_tiled_left:1;
+        bool is_tiled_right:1;
+        enum csd_mode csd_mode;
     } configure;
+
+    int resize_timeout_fd;
 };
 
 struct config;
@@ -423,6 +473,10 @@ struct wayland {
     struct wl_data_device_manager *data_device_manager;
     struct zwp_primary_selection_device_manager_v1 *primary_selection_device_manager;
 
+#if defined(HAVE_XDG_ACTIVATION)
+    struct xdg_activation_v1 *xdg_activation;
+#endif
+
     struct wp_presentation *presentation;
     uint32_t presentation_clock_id;
 
@@ -440,10 +494,19 @@ struct wayland {
 struct wayland *wayl_init(const struct config *conf, struct fdm *fdm);
 void wayl_destroy(struct wayland *wayl);
 
+bool wayl_reload_xcursor_theme(struct seat *seat, int new_scale);
+
 void wayl_flush(struct wayland *wayl);
 void wayl_roundtrip(struct wayland *wayl);
 
-struct wl_window *wayl_win_init(struct terminal *term);
+struct wl_window *wayl_win_init(struct terminal *term, const char *token);
 void wayl_win_destroy(struct wl_window *win);
 
-bool wayl_reload_xcursor_theme(struct seat *seat, int new_scale);
+bool wayl_win_set_urgent(struct wl_window *win);
+
+bool wayl_win_subsurface_new(
+    struct wl_window *win, struct wl_surf_subsurf *surf);
+bool wayl_win_subsurface_new_with_custom_parent(
+    struct wl_window *win, struct wl_surface *parent,
+    struct wl_surf_subsurf *surf);
+void wayl_win_subsurface_destroy(struct wl_surf_subsurf *surf);
